# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Union

import mlrun
from mlrun.artifacts import get_model

from .._common import CommonTypes, ModelHandler


def get_framework_by_instance(model: CommonTypes.ModelType) -> str:
    """
    Get the framework name of the given model by its instance.

    :param model: The model to get his framework.

    :return: The model's framework.

    :raise MLRunInvalidArgumentError: If the given model type is not supported by AutoMLRun or not recognized.
    """
    # TF.Keras:
    try:
        from tensorflow.keras import Model

        from mlrun.frameworks.tf_keras import TFKerasModelHandler

        if isinstance(model, Model):
            return TFKerasModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # PyTorch:
    try:
        from torch.nn import Module

        from mlrun.frameworks.pytorch import PyTorchModelHandler

        if isinstance(model, Module):
            return PyTorchModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # XGBoost:
    try:
        from xgboost import Booster, XGBModel

        from mlrun.frameworks.xgboost import XGBoostModelHandler

        if isinstance(model, (XGBModel, Booster)):
            return XGBoostModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # LightGBM:
    try:
        from lightgbm import Booster, LGBMModel

        from mlrun.frameworks.lgbm import LGBMModelHandler

        if isinstance(model, (LGBMModel, Booster)):
            return LGBMModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # ONNX:
    try:
        from onnx import ModelProto

        from mlrun.frameworks.onnx import ONNXModelHandler

        if isinstance(model, ModelProto):
            return ONNXModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # SciKit-Learn: (As SKLearn's API is commonly used in other frameworks, its important to check it last)
    try:
        import inspect

        import sklearn.base as sklearn_base

        from mlrun.frameworks.sklearn import SKLearnModelHandler

        if isinstance(
            model,
            tuple(
                [
                    class_tuple[1]
                    for class_tuple in inspect.getmembers(sklearn_base, inspect.isclass)
                    if "sklearn" in str(class_tuple[1])
                ]
            ),
        ):
            return SKLearnModelHandler.FRAMEWORK_NAME
    except ModuleNotFoundError:
        pass

    # Unrecognized:
    raise mlrun.errors.MLRunInvalidArgumentError(
        f"The type of the model: {type(model)} was not recognized to be of any of the supported frameworks in "
        f"MLRun or perhaps the required package of the framework could not be imported. If the model's framework is "
        f"one of: TF.Keras, PyTorch, XGBoost, LightGBM, SciKit-Learn, please provide the 'framework' parameter to the "
        f"function."
    )


def get_framework_by_class_name(model: CommonTypes.ModelType) -> str:
    """
    Get the framework name of the given model by its class name.

    :param model: The model to get its framework.

    :return: The model's framework.

    :raise MLRunInvalidArgumentError: If the given model's class name is not supported by AutoMLRun or not recognized.
    """
    # Read the class name:
    class_name = str(model.__class__)

    # Look for the correct framework:
    if "tensorflow" in class_name:
        from mlrun.frameworks.tf_keras import TFKerasModelHandler

        return TFKerasModelHandler.FRAMEWORK_NAME
    if "torch" in class_name:
        from mlrun.frameworks.pytorch import PyTorchModelHandler

        return PyTorchModelHandler.FRAMEWORK_NAME
    if "xgboost" in class_name:
        from mlrun.frameworks.xgboost import XGBoostModelHandler

        return XGBoostModelHandler.FRAMEWORK_NAME
    if "lightgbm" in class_name:
        from mlrun.frameworks.lgbm import LGBMModelHandler

        return LGBMModelHandler.FRAMEWORK_NAME
    if "onnx" in class_name:
        from mlrun.frameworks.onnx import ONNXModelHandler

        return ONNXModelHandler.FRAMEWORK_NAME
    # As SKLearn's API is commonly used in other frameworks, its important to check it last:
    if "sklearn" in class_name:
        from mlrun.frameworks.sklearn import SKLearnModelHandler

        return SKLearnModelHandler.FRAMEWORK_NAME

    # Framework was not recognized:
    raise mlrun.errors.MLRunInvalidArgumentError(
        f"The model's class name: {class_name} was not recognized to be of any of the supported frameworks in "
        f"MLRun. If the model's framework is one of: TF.Keras, PyTorch, XGBoost, LightGBM, SciKit-Learn, "
        f"please provide the 'framework' parameter to the function."
    )


def framework_to_model_handler(framework: str) -> type[ModelHandler]:
    """
    Get the ModelHandler class of the given framework's name.

    :param framework: The framework's name.

    :return: The framework's ModelHandler class.

    :raise MLRunInvalidArgumentError: If the given framework is not supported by AutoMLRun.
    """
    # Match the framework:
    if framework == "tensorflow.keras":
        from mlrun.frameworks.tf_keras import TFKerasModelHandler

        return TFKerasModelHandler
    if framework == "torch":
        from mlrun.frameworks.pytorch import PyTorchModelHandler

        return PyTorchModelHandler
    if framework == "sklearn":
        from mlrun.frameworks.sklearn import SKLearnModelHandler

        return SKLearnModelHandler
    if framework == "xgboost":
        from mlrun.frameworks.xgboost import XGBoostModelHandler

        return XGBoostModelHandler
    if framework == "lightgbm":
        from mlrun.frameworks.lgbm import LGBMModelHandler

        return LGBMModelHandler
    if framework == "onnx":
        from mlrun.frameworks.onnx import ONNXModelHandler

        return ONNXModelHandler

    # Framework was not recognized:
    raise mlrun.errors.MLRunInvalidArgumentError(
        f"The framework {framework} is not supported yet by AutoMLRun. "
        f"Please use the ModelHandler class of the framework from mlrun.frameworks.{framework}"
    )


def framework_to_apply_mlrun(framework: str) -> Callable[..., ModelHandler]:
    """
    Get the 'apply_mlrun' shortcut function of the given framework's name.

    :param framework: The framework's name.

    :return: The framework's 'apply_mlrun' shortcut function.

    :raise MLRunInvalidArgumentError: If the given framework is not supported by AutoMLRun or if it does not have an
                                      'apply_mlrun' yet.
    """
    # Match the framework:
    if framework == "tensorflow.keras":
        from mlrun.frameworks.tf_keras import apply_mlrun

        return apply_mlrun
    if framework == "torch":
        raise mlrun.errors.MLRunInvalidArgumentError(
            "PyTorch has no 'apply_mlrun' shortcut yet. Please use the PyTorchMLRunInterface "
            "from mlrun.frameworks.pytorch instead."
        )
    if framework == "sklearn":
        from mlrun.frameworks.sklearn import apply_mlrun

        return apply_mlrun
    if framework == "xgboost":
        from mlrun.frameworks.xgboost import apply_mlrun

        return apply_mlrun
    if framework == "lightgbm":
        from mlrun.frameworks.lgbm import apply_mlrun

        return apply_mlrun
    if framework == "onnx":
        raise mlrun.errors.MLRunInvalidArgumentError(
            "ONNX has no 'apply_mlrun' shortcut yet. Please use the ONNXMLRunInterface from mlrun.frameworks.onnx "
            "instead."
        )

    # Framework was not recognized:
    raise mlrun.errors.MLRunInvalidArgumentError(
        f"The framework {framework} is not supported yet by AutoMLRun. "
        f"Please use the ModelHandler class of the framework from mlrun.frameworks.{framework}"
    )


class AutoMLRun:
    """
    A library of automatic functions for managing models using MLRun's frameworks package.
    """

    @staticmethod
    def _get_framework(
        model: CommonTypes.ModelType = None, model_path: Optional[str] = None
    ) -> Union[tuple[str, dict]]:
        """
        Try to get the framework from the model or model path provided. The framework can be read from the model path
        only if the model path is of a logged model artifact (store object uri).

        :param model:      The model instance to get its framework.
        :param model_path: The store object uri of a model artifact.

        :return: A tuple of:
                 [0] = The framework.
                 [1] = Collected data (model_file, model_artifact, extra_data) if read from model path or an empty
                       dictionary if read from model.

        :raise MLRunInvalidArgumentError: If both values were None, the framework is not recognized / supported or the
                                          the given model path is not of a store object.
        """
        # If the model is provided:
        if model is not None:
            try:
                # Try to get the framework by the class name:
                return get_framework_by_class_name(model=model), {}
            except mlrun.errors.MLRunInvalidArgumentError:
                # Try to get the framework by the instance's type:
                return get_framework_by_instance(model=model), {}

        # If the model path is provided:
        if model_path is not None:
            # Get the model primary file, extra data and artifact:
            model_file, model_artifact, extra_data = get_model(model_path)
            # If the model path is not of a store object, `model_artifact` will be None:
            if model_artifact is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The model path provided: '{model_path}' is not of a model artifact (store uri) so the framework "
                    f"attribute must be specified."
                )
            elif "framework" not in model_artifact.labels:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"No framework defined for the provided model: '{model_path}'. Please log the model again with "
                    f"the required framework."
                )
            # Return the framework and the collected files and artifacts:
            return (
                model_artifact.labels["framework"],
                {
                    "model_file": model_file,
                    "model_artifact": model_artifact,
                    "extra_data": extra_data,
                },
            )

        # Both values were None:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "At least 'mode' or 'model_path' must be given in order to use AutoMLRun."
        )

    @staticmethod
    def load_model(
        model_path: str,
        model_name: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        framework: Optional[str] = None,
        **kwargs,
    ) -> ModelHandler:
        """
        Load a model using MLRun's ModelHandler. The loaded model can be accessed from the model handler returned
        via model_handler.model. If the model is a store object uri (it is logged in MLRun) then the framework will be
        read automatically, otherwise (for local path and urls) it must be given. The other parameters will be
        automatically read in case its a logged model in MLRun.

        :param model_path:               A store object path of a logged model object in MLRun.
        :param model_name:               The model name to use for storing the model artifact. If not given will have a
                                         default name according to the framework.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key is a
                                         path to a module and its value is the object name to import from it. All the
                                         modules will be imported globally. If multiple objects needed to be imported
                                         from the same module a list can be given. The map can be passed as a path to a
                                         json file as well. For example:

                                         .. code-block:: python

                                             {
                                                 "module1": None,  # import module1
                                                 "module2": ["func1", "func2"],  # from module2 import func1, func2
                                                 "module3.sub_module": "func3",  # from module3.sub_module import func3
                                             }

                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:

                                         .. code-block:: python

                                             {
                                                 "/.../custom_model.py": "MyModel",
                                                 "/.../custom_objects.py": ["object1", "object2"],
                                             }

                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  A MLRun context.
        :param framework:                The model's framework. It must be provided for local paths or urls. If None,
                                         AutoMLRun will assume the model path is of a store uri model artifact and try
                                         to get the framework from it. Default: None.
        :param kwargs:                   Additional parameters for the specific framework's ModelHandler class.

        :return: The model inside a MLRun model handler.

        :raise MLRunInvalidArgumentError: In case the framework is incorrect or missing.
        """
        # Get the model's framework (if needed):
        collected_model_file_and_artifact = {}
        if framework is None:
            framework, collected_model_file_and_artifact = AutoMLRun._get_framework(
                model_path=model_path
            )

        # Get the ModelHandler according to the framework:
        model_handler_class = framework_to_model_handler(framework=framework)

        # Initialize the model handler:
        model_handler = model_handler_class(
            model_path=model_path,
            model_name=model_name,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            **collected_model_file_and_artifact,
            **kwargs,
        )

        # Load the model and return the handler:
        model_handler.load()
        return model_handler

    @staticmethod
    def apply_mlrun(
        model: CommonTypes.ModelType = None,
        model_name: Optional[str] = None,
        tag: str = "",
        model_path: Optional[str] = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        framework: Optional[str] = None,
        auto_log: bool = True,
        **kwargs,
    ) -> ModelHandler:
        """
        Use MLRun's 'apply_mlrun' of the detected given model's framework to wrap the framework relevant methods and
        gain the framework's features in MLRun. A ModelHandler initialized with the model will be returned.

        :param model:                    The model to wrap. Can be loaded from the model path given as well.
        :param model_name:               The model name to use for storing the model artifact. If not given will have a
                                         default name according to the framework.
        :param tag:                      The model's tag to log with.
        :param model_path:               The model's store object path. Mandatory for evaluation (to know which model to
                                         update). If model is not provided, it will be loaded from this path.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key is a
                                         path to a module and its value is the object name to import from it. All the
                                         modules will be imported globally. If multiple objects needed to be imported
                                         from the same module a list can be given. The map can be passed as a path to a
                                         json file as well. For example:

                                         .. code-block:: python

                                             {
                                                 "module1": None,  # import module1
                                                 "module2": ["func1", "func2"],  # from module2 import func1, func2
                                                 "module3.sub_module": "func3",  # from module3.sub_module import func3
                                             }

                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:

                                         .. code-block:: python

                                             {
                                                 "/.../custom_model.py": "MyModel",
                                                 "/.../custom_objects.py": ["object1", "object2"],
                                             }

                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  A MLRun context.
        :param auto_log:                 Whether to enable auto-logging capabilities of MLRun or not. Auto logging will
                                         add default artifacts and metrics besides the one you can pass here.
        :param framework:                The model's framework. If None, AutoMLRun will try to figure out the framework.
                                         From the provided model or model path. Default: None.
        :param kwargs:                   Additional parameters for the specific framework's 'apply_mlrun' function like
                                         metrics, callbacks and more (read the docs of the required framework to know
                                         more).

        :return: The framework's model handler initialized with the given model.
        """
        # Get the model's framework:
        collected_model_file_and_artifact = {}
        if framework is None:
            framework, collected_model_file_and_artifact = AutoMLRun._get_framework(
                model=model, model_path=model_path
            )

        # Get the framework's 'apply_mlrun':
        apply_mlrun = framework_to_apply_mlrun(framework=framework)

        # Initialize the model handler:
        return apply_mlrun(
            model=model,
            model_name=model_name,
            tag=tag,
            model_path=model_path,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            auto_log=auto_log,
            model_handler_kwargs=collected_model_file_and_artifact,
            **kwargs,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from time import sleep

import yaml
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.jobs import RunLifeCycleState
from databricks_wrapper import credentials_path

from mlrun.errors import MLRunRuntimeError


def main():
    if not os.path.exists(credentials_path):
        print("credentials_path not found. Retrying in 10 seconds...")
        sleep(10)
        if not os.path.exists(credentials_path):
            raise MLRunRuntimeError(
                "The Databricks credentials path does not exist."
                " Please manually cancel the job from the Databricks environment."
            )
    with open(credentials_path) as yaml_file:
        loaded_data = yaml.safe_load(yaml_file)
    # use for flat yaml only
    for key, value in loaded_data.items():
        os.environ[key] = str(value)
    workspace = WorkspaceClient()
    task_id = os.environ.get("TASK_RUN_ID")
    if os.environ.get("IS_FINISHED", "false").lower() == "false":
        run = workspace.jobs.cancel_run(run_id=task_id).result()
        life_cycle_state = run.as_dict().get("state").get("life_cycle_state")
        if (
            # TERMINATED is also the life_cycle_state of tasks that have already either failed or succeeded
            life_cycle_state
            not in [RunLifeCycleState.TERMINATING, RunLifeCycleState.TERMINATED]
        ):
            raise MLRunRuntimeError(
                f"Cancelling task {task_id} has failed, life cycle state is: {life_cycle_state}. "
                f"Please check the status of this task in the Databricks environment."
            )
        print(f"Cancelling task {task_id} has succeeded.")


main()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import json
from abc import ABC, abstractmethod

from pydantic.v1 import validator
from pydantic.v1.dataclasses import dataclass

import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.errors
from mlrun.utils import logger

_RESULT_EXTRA_DATA_MAX_SIZE = 998


class _ModelMonitoringApplicationDataRes(ABC):
    name: str

    def __post_init__(self):
        if not mm_constants.RESULT_NAME_REGEX.fullmatch(self.name):
            raise mlrun.errors.MLRunValueError(
                "The application result or metric name must comply with the regex "
                f"`{mm_constants.RESULT_NAME_REGEX.pattern}`"
            )

    @abstractmethod
    def to_dict(self):
        raise NotImplementedError


@dataclass
class ModelMonitoringApplicationResult(_ModelMonitoringApplicationDataRes):
    """
    Class representing the result of a custom model monitoring application.

    :param name:           (str) Name of the application result. This name must be
                            unique for each metric in a single application
                            (name must be of the format :code:`[a-zA-Z_][a-zA-Z0-9_]*`).
    :param value:          (float) Value of the application result.
    :param kind:           (ResultKindApp) Kind of application result.
    :param status:         (ResultStatusApp) Status of the application result.
    :param extra_data:     (dict) Extra data associated with the application result. Note that if the extra data is
                                  exceeding the maximum size of 998 characters, it will be ignored and a message will
                                  be logged. In this case, we recommend logging the extra data as a separate artifact or
                                  shortening it.
    """

    name: str
    value: float
    kind: mm_constants.ResultKindApp
    status: mm_constants.ResultStatusApp
    extra_data: dict = dataclasses.field(default_factory=dict)

    def to_dict(self):
        """
        Convert the object to a dictionary format suitable for writing.

        :returns:    (dict) Dictionary representation of the result.
        """
        return {
            mm_constants.ResultData.RESULT_NAME: self.name,
            mm_constants.ResultData.RESULT_VALUE: self.value,
            mm_constants.ResultData.RESULT_KIND: self.kind.value,
            mm_constants.ResultData.RESULT_STATUS: self.status.value,
            mm_constants.ResultData.RESULT_EXTRA_DATA: json.dumps(self.extra_data),
        }

    @validator("extra_data")
    @classmethod
    def validate_extra_data_len(cls, result_extra_data: dict):
        """Ensure that the extra data is not exceeding the maximum size which is important to avoid
        possible storage issues."""
        extra_data_len = len(json.dumps(result_extra_data))
        if extra_data_len > _RESULT_EXTRA_DATA_MAX_SIZE:
            logger.warning(
                f"Extra data is too long and won't be stored: {extra_data_len} characters while the maximum "
                f"is {_RESULT_EXTRA_DATA_MAX_SIZE} characters."
                f"Please shorten the extra data or log it as a separate artifact."
            )
            return {}
        return result_extra_data


@dataclass
class ModelMonitoringApplicationMetric(_ModelMonitoringApplicationDataRes):
    """
    Class representing a single metric of a custom model monitoring application.

    :param name:           (str) Name of the application metric. This name must be
                            unique for each metric in a single application
                            (name must be of the format :code:`[a-zA-Z_][a-zA-Z0-9_]*`).
    :param value:          (float) Value of the application metric.
    """

    name: str
    value: float

    def to_dict(self):
        """
        Convert the object to a dictionary format suitable for writing.

        :returns:    (dict) Dictionary representation of the result.
        """
        return {
            mm_constants.MetricData.METRIC_NAME: self.name,
            mm_constants.MetricData.METRIC_VALUE: self.value,
        }


@dataclasses.dataclass
class _ModelMonitoringApplicationStats(_ModelMonitoringApplicationDataRes):
    """
    Class representing the stats of histogram data drift application.

    :param name             (mm_constant.StatsKind) Enum mm_constant.StatsData of the stats data kind of the event
    :param                  (str) iso format representation of the timestamp the event took place
    :param stats            (dict) Dictionary representation of the stats calculated for the event

    """

    name: mm_constants.StatsKind
    timestamp: str
    stats: dict = dataclasses.field(default_factory=dict)

    def to_dict(self):
        """
        Convert the object to a dictionary format suitable for writing.

        :returns:    (dict) Dictionary representation of the result.
        """
        return {
            mm_constants.StatsData.STATS_NAME: self.name,
            mm_constants.StatsData.STATS: self.stats,
            mm_constants.StatsData.TIMESTAMP: self.timestamp,
        }



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

import numpy as np
import transformers

import mlrun
from mlrun.serving.v2_serving import V2ModelServer


class HuggingFaceModelServer(V2ModelServer):
    """
    Hugging Face Model serving class, inheriting the V2ModelServer class for being initialized automatically by the
    model server and be able to run locally as part of a nuclio serverless function, or as part of a real-time pipeline.
    Notice:
        In order to use this serving class, please ensure that the transformers package is installed.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        name: Optional[str] = None,
        task: Optional[str] = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        model_class: Optional[str] = None,
        tokenizer_name: Optional[str] = None,
        tokenizer_class: Optional[str] = None,
        framework: Optional[str] = None,
        **class_args,
    ):
        """
        Initialize a serving class for a Hugging face model.
        :param context:         For internal use (passed in init).
        :param name:            The name of this server to be initialized
        :param task:            The task defining which pipeline will be returned.
        :param model_path:      Not in use. When adding a model pass any string value
        :param model_name:      The model's name in the Hugging Face hub
                                e.g., `nlptown/bert-base-multilingual-uncased-sentiment`
        :param model_class:     The model class type object which can be passed as the class's name (string).
                                Must be provided and to be matched with `model_name`.
                                e.g., `AutoModelForSequenceClassification`
        :param tokenizer_name:  The name of the tokenizer in the Hugging Face hub
                                e.g., `nlptown/bert-base-multilingual-uncased-sentiment`
        :param tokenizer_class: The model's class type object which can be passed as the class's name (string).
                                Must be provided and to be matched with `model_name`.
                                e.g., `AutoTokenizer`
        :param framework:       The framework to use, either `"pt"` for PyTorch or `"tf"` for TensorFlow. The specified
                                framework must be installed.
                                If no framework is specified, will default to the one currently installed.
                                If no framework is specified and both frameworks are installed, will default to the
                                framework of the `model`, or to PyTorch if no model is provided
        :param class_args:      -
        """
        super().__init__(
            context=context,
            name=name,
            model_path=model_path,
            **class_args,
        )
        self.task = task
        self.model = None
        self.tokenizer = None
        self.model_name = model_name
        self.tokenizer_name = tokenizer_name
        self.model_class = model_class
        self.tokenizer_class = tokenizer_class
        self.framework = framework
        self.pipe = None

    def load(self):
        """
        loads the model and the tokenizer and initializes the pipeline based on them.
        """

        # Loading the pretrained model:
        if self.model_class:
            model_object = getattr(transformers, self.model_class)
            self.model = model_object.from_pretrained(self.model_name)

        # Loading the pretrained tokenizer:
        if self.tokenizer_class:
            tokenizer_object = getattr(transformers, self.tokenizer_class)
            self.tokenizer = tokenizer_object.from_pretrained(self.tokenizer_name)

        # Initializing the pipeline:
        self.pipe = transformers.pipeline(
            task=self.task,
            model=self.model or self.model_name,
            tokenizer=self.tokenizer,
            framework=self.framework,
        )

    def predict(self, request: dict[str, Any]) -> list:
        """
        Generate model predictions from sample.
        :param request: The request to the model. The input to the model will be read from the "inputs" key.
        :return: The model's prediction on the given input.
        """
        # Get the inputs:
        inputs = request["inputs"]

        # Applying prediction according the inputs shape:
        result = (
            [self.pipe(**_input) for _input in inputs]
            if isinstance(inputs[0], dict)
            else self.pipe(inputs)
        )

        # Arranging the result into a List[Dict]
        # (This is necessary because the result may vary from one model to another)
        if all(isinstance(res, list) for res in result):
            result = [res[0] for res in result]

        # Converting JSON non-serializable numpy objects to native types:
        for res in result:
            for key, val in res.items():
                if isinstance(val, np.generic):
                    res[key] = val.item()
                elif isinstance(val, np.ndarray):
                    res[key] = val.tolist()

        return result

    def explain(self, request: dict) -> str:
        """
        Return a string explaining what model is being served in this serving function and the function name.
        :param request: A given request.
        :return: Explanation string.
        """
        return f"The '{self.model_name}' model serving function named '{self.name}"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from types import ModuleType
from typing import Callable, Optional, Union

import lightgbm as lgb

import mlrun

from ..._common import MLRunInterface
from ..._ml_common import MLArtifactsLibrary, MLTypes
from ..callbacks import Callback, MLRunLoggingCallback
from ..model_handler import LGBMModelHandler
from ..utils import LGBMTypes, LGBMUtils
from .booster_mlrun_interface import LGBMBoosterMLRunInterface


class LGBMMLRunInterface(MLRunInterface, ABC):
    """
    Interface for adding MLRun features for the LightGBM module (Training API).
    """

    # Attributes to be inserted so the MLRun interface will be fully enabled.
    _PROPERTIES = {
        # A list for collecting all the mlrun API callbacks (collected automatically):
        "_mlrun_callbacks": [],  # type: List[Callback]
        # Context to use for logging:
        "_context": None,  # type: mlrun.MLClientCtx
        # Model logging configurations:
        "_log_model": False,
        "_model_handler_kwargs": {},
        "_log_model_kwargs": {},  # Keyword arguments for the model handler's log method.
        # Training logging configurations:
        "_log_training": False,
        "_mlrun_logging_callback_kwargs": {},
    }
    _FUNCTIONS = [
        "configure_logging",
        "_parse_parameters",
        "_parse_callbacks",
        "_pre_train",
        "_post_train",
    ]

    # Attributes to replace so the MLRun interface will be fully enabled.
    _REPLACED_FUNCTIONS = [
        "train",
        # TODO: Wrap `cv` as well.
    ]

    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-lightgbm"

    @classmethod
    def add_interface(
        cls,
        obj: Optional[ModuleType] = None,
        restoration: LGBMTypes.MLRunInterfaceRestorationType = None,
    ):
        """
        Enrich the object with this interface properties, methods and functions, so it will have LightGBM MLRun's
        features.

        :param obj:         The object to enrich his interface.
        :param restoration: Restoration information tuple as returned from 'remove_interface' in order to add the
                            interface in a certain state.
        """
        # If the lightgbm module was not provided:
        if obj is None:
            # Set it to the module imported here:
            obj = lgb
            # See if lightgbm was imported outside this file's scope:
            if all(lgb_import not in globals() for lgb_import in ["lightgbm", "lgb"]):
                # Import lightgbm globally:
                globals().update({"lightgbm": lgb, "lgb": lgb})

        # Add the interface to the provided lightgbm module:
        super().add_interface(obj=obj, restoration=restoration)

    @staticmethod
    def mlrun_train(*args, **kwargs):
        """
        MLRun's `lightgbm.train` wrapper. It will parse the parameters and run the training supervised by MLRun.
        """
        # Get the training parameters (must be given) and parse them:
        parameters, _ = MLRunInterface._get_function_argument(
            func=lgb.original_train,
            argument_name="params",
            passed_args=args,
            passed_kwargs=kwargs,
            default_value={},
        )
        lgb._parse_parameters(parameters=parameters)

        # Get the training set (must be given):
        train_set, _ = MLRunInterface._get_function_argument(
            func=lgb.original_train,
            argument_name="train_set",
            passed_args=args,
            passed_kwargs=kwargs,
            default_value=None,
        )
        train_set = (  # Post training LightGBBM is consuming the data, so we keep it in a tuple.
            train_set.data,
            train_set.label,
        )

        # Get the validation sets:
        validation_sets, _ = MLRunInterface._get_function_argument(
            func=lgb.original_train,
            argument_name="valid_sets",
            passed_args=args,
            passed_kwargs=kwargs,
            default_value=[],
        )
        validation_sets_names, _ = MLRunInterface._get_function_argument(
            func=lgb.original_train,
            argument_name="valid_names",
            passed_args=args,
            passed_kwargs=kwargs,
            default_value=[f"valid_{i}" for i in range(len(validation_sets))],
        )
        validation_sets = [
            ((validation_set.data, validation_set.label), name)
            for validation_set, name in zip(validation_sets, validation_sets_names)
        ]

        # Collect the mlrun callbacks from the provided callbacks:
        callbacks, is_given = MLRunInterface._get_function_argument(
            func=lgb.original_train,
            argument_name="callbacks",
            passed_args=args,
            passed_kwargs=kwargs,
            default_value=[],
        )
        lgb._parse_callbacks(callbacks=callbacks)
        if is_given is None:
            kwargs["callbacks"] = callbacks

        # Call the pre-train function:
        lgb._pre_train()

        # Call the original train function:
        booster = lgb.original_train(*args, **kwargs)

        # Call the post train function:
        lgb._post_train(
            booster=booster, train_set=train_set, validation_sets=validation_sets
        )

        return booster

    @staticmethod
    def configure_logging(
        context: mlrun.MLClientCtx = None,
        log_model: bool = True,
        model_handler_kwargs: Optional[dict] = None,
        log_model_kwargs: Optional[dict] = None,
        log_training: bool = True,
        mlrun_logging_callback_kwargs: Optional[dict] = None,
    ):
        """
        Configure the logging of the training API in LightGBM to log the training and model into MLRun. Each `train`
        call will use these configurations to initialize callbacks and log the model at the end of training.

        :param context:                       MLRun context to log to.
        :param log_model:                     Whether to log the model at the end of training. Default: True.
        :param model_handler_kwargs:          Keyword arguments to use for initializing the model handler for the newly
                                              trained model.
        :param log_model_kwargs:              Keyword arguments to use for calling the handler's `log` method.
        :param log_training:                  Whether to log the training metrics and hyperparameters to MLRun.
        :param mlrun_logging_callback_kwargs: Keyword arguments to use for initializing the `MLRunLoggingCallback` for
                                              logging the training results during and post training.

        :raise MLRunInvalidArgumentError: In case the 'model' keyword argument was provided in the
                                          `model_handler_kwargs`.
        """
        # Store the context:
        lgb._context = context

        # Store the given model logging configurations:
        lgb._log_model = log_model
        if model_handler_kwargs is not None:
            if "model" in model_handler_kwargs:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "The 'model' keyword cannot be passed in the model handler keyword arguments as it will be used to "
                    "set the booster that was trained."
                )
            lgb._model_handler_kwargs = model_handler_kwargs
        if log_model_kwargs is not None:
            lgb._log_model_kwargs = log_model_kwargs

        # Store the given training logging configurations:
        lgb._log_training = log_training
        if mlrun_logging_callback_kwargs is not None:
            lgb._mlrun_logging_callback_kwargs = mlrun_logging_callback_kwargs

    @staticmethod
    def _parse_parameters(parameters: dict):
        """
        Parse the parameters passed to the training API functions of LightGBM for enabling MLRun quality of life
        features.

        Notice: The parameters dictionary should be edited and not returned as it is passed by reference.

        :param parameters: The `params` argument that was passed.
        """
        # TODO: Parse `input_model` - if it's set and is a store path, download it locally and set the parameter to the
        #       temp path.
        pass

    @staticmethod
    def _parse_callbacks(callbacks: list[Callable]):
        """
        Parse the callbacks passed to the training API functions of LightGBM for adding logging and enabling the MLRun
        callbacks API.

        Notice: The callbacks list should be edited and not returned as it is passed by reference.

        :param callbacks: The `callbacks` argument that was passed.
        """
        # Check if needed to log training:
        if lgb._log_training:
            # Add only if it was not given already by the user:
            if MLRunLoggingCallback not in [type(callback) for callback in callbacks]:
                callbacks.append(
                    MLRunLoggingCallback(
                        context=lgb._context, **lgb._mlrun_logging_callback_kwargs
                    )
                )

        # Collect all the MLRun API callbacks:
        lgb._mlrun_callbacks.clear()
        for callback in callbacks:
            if isinstance(callback, Callback):
                lgb._mlrun_callbacks.append(callback)

    @staticmethod
    def _pre_train():
        """
        Called pre-training to call the mlrun callbacks `on_train_begin` method.
        """
        for callback in lgb._mlrun_callbacks:
            callback.on_train_begin()

    @staticmethod
    def _post_train(
        booster: lgb.Booster,
        train_set: tuple[MLTypes.DatasetType, Union[MLTypes.DatasetType, None]],
        validation_sets: list[
            tuple[tuple[MLTypes.DatasetType, Union[MLTypes.DatasetType, None]], str]
        ],
    ):
        """
        Called post training to call the mlrun callbacks `on_train_end` method and to log the model.

        :param booster:         The booster to log.
        :param train_set:       The training set that was used to train the given booster.
        :param validation_sets: The validation sets used for training validation.
        """
        # Call the `on_train_end` method of the callbacks while collecting extra data from the mlrun logging callback:
        extra_data = {}
        metrics = {}
        for callback in lgb._mlrun_callbacks:
            callback.on_train_end()
            if isinstance(callback, MLRunLoggingCallback):
                extra_data = {**extra_data, **callback.logger.get_artifacts()}
                metrics = {**metrics, **callback.logger.get_metrics()}

        # Try to produce validation artifacts:  # TODO: Temporary until the Evaluator class will be implemented!
        algorithm_functionality = LGBMUtils.get_algorithm_functionality(
            objective=booster.params.get("objective", "regression")
        )
        plans = [MLArtifactsLibrary.feature_importance()]
        if algorithm_functionality.is_classification():
            plans += [
                MLArtifactsLibrary.calibration_curve(),
                MLArtifactsLibrary.roc_curve(),
                MLArtifactsLibrary.confusion_matrix(),
            ]
        validations_artifacts = {}
        for validation_set, validation_set_name in validation_sets:
            artifacts = {}
            y_pred = booster.predict(validation_set[0])
            plans.append(MLArtifactsLibrary.dataset(name=validation_set_name))
            for plan in plans:
                try:
                    artifacts = {
                        **artifacts,
                        **plan.produce(
                            model=booster,
                            x=validation_set[0],
                            y=validation_set[1],
                            y_pred=y_pred,
                        ),
                    }
                except Exception:
                    pass
            artifacts = {
                f"{validation_set_name}-{key}": value
                for key, value in artifacts.items()
            }
            for artifact in artifacts.values():
                if validation_set_name not in artifact.key:
                    artifact.metadata.key = f"{validation_set_name}-{artifact.key}"
                lgb._context.log_artifact(artifact)
            validations_artifacts = {**validations_artifacts, **artifacts}
            plans.pop(-1)

        # Apply the booster MLRun interface:
        LGBMBoosterMLRunInterface.add_interface(obj=booster)

        # Set the handler to the booster:
        booster.model_handler = LGBMModelHandler(
            model=booster, context=lgb._context, **lgb._model_handler_kwargs
        )

        # Register found extra data and metrics:
        booster.model_handler.set_extra_data(to_add=extra_data)
        booster.model_handler.set_metrics(to_add=metrics)

        # Set the sample set to the training set if None:
        if lgb._log_model_kwargs.get("sample_set", None) is None:
            sample_set, target_columns = LGBMUtils.concatenate_x_y(
                x=train_set[0],
                y=train_set[1],
                target_columns_names=lgb._log_model_kwargs.get("target_columns", None),
            )
            booster.model_handler.set_target_columns(target_columns=target_columns)
            booster.model_handler.set_sample_set(sample_set=sample_set)

        # Check if needed to log the model:
        if lgb._log_model:
            booster.model_handler.register_artifacts(artifacts=validations_artifacts)
            booster.model_handler.log(**lgb._log_model_kwargs)

        lgb._context.commit(completed=False)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import typing
from datetime import datetime

import numpy as np
import pandas as pd
from deprecated import deprecated

import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.datastore.base
import mlrun.feature_store
import mlrun.model_monitoring.applications as mm_app
import mlrun.serving
from mlrun.common.schemas import ModelEndpoint
from mlrun.common.schemas.model_monitoring import (
    FunctionURI,
)
from mlrun.data_types.infer import InferOptions, get_df_stats
from mlrun.utils import check_if_hub_uri, datetime_now, logger, merge_requirements

from ..common.schemas.hub import HubModuleType
from .helpers import update_model_endpoint_last_request

# A union of all supported dataset types:
DatasetType = typing.Union[
    mlrun.datastore.base.DataItem,
    list,
    dict,
    pd.DataFrame,
    pd.Series,
    np.ndarray,
    typing.Any,
]


# TODO: Remove this in 1.12.0
@deprecated(
    version="1.10.0",
    reason="This function is deprecated and will be removed in 1.12. You can generate a model endpoint by either "
    "deploying a monitored serving function as a real-time service or running it as an offline job. "
    "To retrieve model endpoints, use `project.list_model_endpoints()`",
    category=FutureWarning,
)
def get_or_create_model_endpoint(
    project: str,
    model_endpoint_name: str,
    model_path: str = "",
    endpoint_id: str = "",
    function_name: str = "",
    function_tag: str = "latest",
    context: typing.Optional["mlrun.MLClientCtx"] = None,
    sample_set_statistics: typing.Optional[dict[str, typing.Any]] = None,
    monitoring_mode: mm_constants.ModelMonitoringMode = mm_constants.ModelMonitoringMode.enabled,
    db_session=None,
    feature_analysis: bool = False,
) -> ModelEndpoint:
    """
    Get a single model endpoint object. If not exist, generate a new model endpoint with the provided parameters. Note
    that in case of generating a new model endpoint, by default the monitoring features are disabled. To enable these
    features, set `monitoring_mode=enabled`.

    :param project:                  Project name.
    :param model_endpoint_name:      If a new model endpoint is created, the model endpoint name will be presented
                                     under this endpoint (applicable only to new endpoint_id).
    :param model_path:               The model store path (applicable only to new endpoint_id).
    :param endpoint_id:              Model endpoint unique ID. If not exist in DB, will generate a new record with a
                                     newly generated ID.
    :param function_name:            If a new model endpoint is created, use this function name.
    :param function_tag:             If a new model endpoint is created, use this function tag.
    :param context:                  MLRun context. If `function_name` not provided, use the context to generate the
                                     full function hash.
    :param sample_set_statistics:    Dictionary of sample set statistics that will be used as a reference data for
                                     the new model endpoint (applicable only to new endpoint_id).
    :param monitoring_mode:          If enabled, apply model monitoring features on the provided endpoint id
                                     (applicable only to new endpoint_id).
    :param db_session:               A runtime session that manages the current dialog with the database.
    :param feature_analysis:         If True, the model endpoint will be retrieved with the feature analysis mode.

    :return: A ModelEndpoint object
    """

    if not db_session:
        # Generate a runtime database
        db_session = mlrun.get_run_db()
    model_endpoint = None
    if not function_name and context:
        function_name = FunctionURI.from_string(
            context.to_dict()["spec"]["function"]
        ).function
    if endpoint_id or function_name:
        try:
            model_endpoint = db_session.get_model_endpoint(
                project=project,
                name=model_endpoint_name,
                endpoint_id=endpoint_id,
                function_name=function_name,
                function_tag=function_tag or "latest",
                feature_analysis=feature_analysis,
            )
            # If other fields provided, validate that they are correspond to the existing model endpoint data
            _model_endpoint_validations(
                model_endpoint=model_endpoint,
                model_path=model_path,
                sample_set_statistics=sample_set_statistics,
            )

        except mlrun.errors.MLRunNotFoundError:
            # Create a new model endpoint with the provided details
            pass
    if not model_endpoint:
        model_endpoint = _generate_model_endpoint(
            project=project,
            db_session=db_session,
            model_path=model_path,
            model_endpoint_name=model_endpoint_name,
            function_name=function_name,
            function_tag=function_tag,
            monitoring_mode=monitoring_mode,
        )
    return model_endpoint


# TODO: Remove this in 1.12.0
@deprecated(
    version="1.10.0",
    reason="This function is deprecated and will be removed in 1.12. "
    "Instead, run a monitored serving function as a job",
    category=FutureWarning,
)
def record_results(
    project: str,
    model_path: str,
    model_endpoint_name: str,
    endpoint_id: str = "",
    function_name: str = "",
    context: typing.Optional["mlrun.MLClientCtx"] = None,
    infer_results_df: typing.Optional[pd.DataFrame] = None,
    sample_set_statistics: typing.Optional[dict[str, typing.Any]] = None,
    monitoring_mode: mm_constants.ModelMonitoringMode = mm_constants.ModelMonitoringMode.enabled,
) -> ModelEndpoint:
    """
    Write a provided inference dataset to model endpoint parquet target. If not exist, generate a new model endpoint
    record and use the provided sample set statistics as feature stats that will be used later for the drift analysis.
    To activate model monitoring, run `project.enable_model_monitoring()`. The model monitoring applications will be
    triggered with the recorded data according to a periodic schedule.

    :param project:                  Project name.
    :param model_path:               The model Store path.
    :param model_endpoint_name:      If a new model endpoint is generated, the model endpoint name will be presented
                                     under this endpoint.
    :param endpoint_id:              Model endpoint unique ID. If not exist in DB, will generate a new record with a
                                     newly generated ID.
    :param function_name:            If a new model endpoint is created, use this function name for generating the
                                     function URI.
    :param context:                  MLRun context. Note that the context is required generating the model endpoint.
    :param infer_results_df:         DataFrame that will be stored under the model endpoint parquet target. Will be
                                     used for doing the drift analysis. Please make sure that the dataframe includes
                                     both feature names and label columns. If you are recording results for existing
                                     model endpoint, the endpoint should be a batch endpoint.
    :param sample_set_statistics:    Dictionary of sample set statistics that will be used as a reference data for
                                     the current model endpoint.
    :param monitoring_mode:          If enabled, apply model monitoring features on the provided endpoint id. Enabled
                                     by default.

    :return: A ModelEndpoint object
    """

    db = mlrun.get_run_db()

    model_endpoint = get_or_create_model_endpoint(
        project=project,
        endpoint_id=endpoint_id,
        model_path=model_path,
        model_endpoint_name=model_endpoint_name,
        function_name=function_name,
        context=context,
        sample_set_statistics=sample_set_statistics,
        monitoring_mode=monitoring_mode,
        db_session=db,
    )
    logger.debug("Model endpoint", endpoint=model_endpoint)

    if infer_results_df is not None:
        if (
            model_endpoint.metadata.endpoint_type
            != mlrun.common.schemas.model_monitoring.EndpointType.BATCH_EP
        ):
            logger.warning(
                "Inference results can be recorded only for batch endpoints. "
                "Therefore the current results won't be monitored."
            )
        else:
            timestamp = datetime_now()
            # Write the monitoring parquet to the relevant model endpoint context
            write_monitoring_df(
                feature_set_uri=model_endpoint.spec.monitoring_feature_set_uri,
                infer_datetime=timestamp,
                endpoint_id=model_endpoint.metadata.uid,
                infer_results_df=infer_results_df,
            )

            # Update the last request time
            update_model_endpoint_last_request(
                project=project,
                model_endpoint=model_endpoint,
                current_request=timestamp,
                db=db,
            )

    return model_endpoint


def _model_endpoint_validations(
    model_endpoint: ModelEndpoint,
    model_path: str = "",
    sample_set_statistics: typing.Optional[dict[str, typing.Any]] = None,
) -> None:
    """
    Validate that provided model endpoint configurations match the stored fields of the provided `ModelEndpoint`
    object. Usually, this method is called by `get_or_create_model_endpoint()` in cases that the model endpoint
    already exist. If one of the validations fails, this method might raise an error, indicating on possible conflict.

    :param model_endpoint:           A `ModelEndpoint` object that contains the expected values.
    :param model_path:               Model store path. In case of endpoint_id reuse, should be similar to the model_uri
                                     that is stored under model_endpoint.spec.model_uri. Model endpoint record refers
                                     to a single model store path.
    :param sample_set_statistics:    Dictionary of sample set statistics. Once the model endpoint is registered, it
                                     is forbidden to provide a different reference data to that model endpoint.
                                     In case of discrepancy between the provided `sample_set_statistics` and the
                                     `model_endpoints.spec.feature_stats`, a warning will be presented to the user.
    """

    # Model Path
    if model_path:
        # Generate the parsed model uri that is based on hash, key, iter, and tree
        model_obj = mlrun.datastore.get_store_resource(model_path)

        model_artifact_uri = mlrun.utils.helpers.generate_artifact_uri(
            project=model_endpoint.metadata.project,
            key=model_obj.key,
            iter=model_obj.iter,
            tree=model_obj.tree,
            uid=model_obj.uid,
        )

        # Enrich the uri schema with the store prefix
        model_artifact_uri = mlrun.datastore.get_store_uri(
            kind=mlrun.utils.helpers.StorePrefix.Model, uri=model_artifact_uri
        )

        if model_endpoint.spec.model_uri != model_artifact_uri:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"provided model store path {model_path} does not match "
                f"the path that is stored under the existing model "
                f"endpoint record: {model_endpoint.spec.model_uri}"
            )

    # Feature stats
    if (
        sample_set_statistics
        and sample_set_statistics != model_endpoint.spec.feature_stats
    ):
        logger.warning(
            "Provided sample set statistics is different from the registered statistics. "
            "If new sample set statistics is to be used, new model endpoint should be created"
        )


def write_monitoring_df(
    endpoint_id: str,
    infer_results_df: pd.DataFrame,
    infer_datetime: datetime,
    monitoring_feature_set: typing.Optional["mlrun.feature_store.FeatureSet"] = None,
    feature_set_uri: str = "",
) -> None:
    """Write infer results dataframe to the monitoring parquet target of the current model endpoint. The dataframe will
    be written using feature set ingest process. Please make sure that you provide either a valid monitoring feature
    set (with parquet target) or a valid monitoring feature set uri.

    :param endpoint_id:             Model endpoint unique ID.
    :param infer_results_df:        DataFrame that will be stored under the model endpoint parquet target.
    :param monitoring_feature_set:  A `mlrun.feature_store.FeatureSet` object corresponding to the provided endpoint_id.
    :param feature_set_uri:         if monitoring_feature_set not provided, use the feature_set_uri value to get the
                                    relevant `mlrun.feature_store.FeatureSet`.
    """

    if not monitoring_feature_set:
        if not feature_set_uri:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Please provide either a valid monitoring feature set object or a monitoring feature set uri"
            )

        monitoring_feature_set = mlrun.feature_store.get_feature_set(
            uri=feature_set_uri
        )

    # Modify the DataFrame to the required structure that will be used later by the monitoring batch job
    if mm_constants.EventFieldType.TIMESTAMP not in infer_results_df.columns:
        # Initialize timestamp column with the current time
        infer_results_df[mm_constants.EventFieldType.TIMESTAMP] = infer_datetime

    # `endpoint_id` is the monitoring feature set entity and therefore it should be defined as the df index before
    # the ingest process
    infer_results_df[mm_constants.EventFieldType.ENDPOINT_ID] = endpoint_id
    infer_results_df.set_index(mm_constants.EventFieldType.ENDPOINT_ID, inplace=True)

    monitoring_feature_set.ingest(source=infer_results_df, overwrite=False)


def _generate_model_endpoint(
    project: str,
    db_session,
    model_path: str,
    model_endpoint_name: str,
    function_name: str,
    function_tag: str,
    monitoring_mode: mm_constants.ModelMonitoringMode = mm_constants.ModelMonitoringMode.enabled,
) -> ModelEndpoint:
    """
    Write a new model endpoint record.

    :param project:                  Project name.

    :param db_session:               A session that manages the current dialog with the database.
    :param model_path:               The model Store path.
    :param model_endpoint_name:      Model endpoint name will be presented under the new model endpoint.
    :param function_name:            If a new model endpoint is created, use this function name.
    :param function_tag:             If a new model endpoint is created, use this function tag.
    :param monitoring_mode:          Monitoring mode of the new model endpoint.

    :return `mlrun.common.schemas.ModelEndpoint` object.
    """

    current_time = datetime_now()
    model_endpoint = mlrun.common.schemas.ModelEndpoint(
        metadata=mlrun.common.schemas.ModelEndpointMetadata(
            project=project,
            name=model_endpoint_name,
            endpoint_type=mlrun.common.schemas.model_monitoring.EndpointType.BATCH_EP,
            # Due to backwards compatibility, this endpoint will be created as a legacy batch endpoint.
            mode=mlrun.common.schemas.model_monitoring.EndpointMode.BATCH_LEGACY,
        ),
        spec=mlrun.common.schemas.ModelEndpointSpec(
            function_name=function_name or "function",
            function_tag=function_tag or "latest",
            model_path=model_path,
            model_class="drift-analysis",
        ),
        status=mlrun.common.schemas.ModelEndpointStatus(
            monitoring_mode=monitoring_mode,
            first_request=current_time,
            last_request=current_time,
        ),
    )

    return db_session.create_model_endpoint(model_endpoint=model_endpoint)


def get_sample_set_statistics(
    sample_set: DatasetType = None,
    model_artifact_feature_stats: typing.Optional[dict] = None,
    sample_set_columns: typing.Optional[list] = None,
    sample_set_drop_columns: typing.Optional[list] = None,
    sample_set_label_columns: typing.Optional[list] = None,
) -> dict:
    """
    Get the sample set statistics either from the given sample set or the statistics logged with the model while
    favoring the given sample set.

    :param sample_set:                   A sample dataset to give to compare the inputs in the drift analysis.
    :param model_artifact_feature_stats: The `feature_stats` attribute in the spec of the model artifact, where the
                                         original sample set statistics of the model was used.
    :param sample_set_columns: The column names of sample_set.
    :param sample_set_drop_columns: ``str`` / ``int`` or a list of ``str`` / ``int`` that
                                    represent the column names / indices to drop.
    :param sample_set_label_columns: The target label(s) of the column(s) in the dataset. for Regression or
                                     Classification tasks.
    :returns: The sample set statistics.

    raises MLRunInvalidArgumentError: If no sample set or statistics were given.
    """
    # Check if a sample set was provided:
    if sample_set is None:
        # Check if the model was logged with a sample set:
        if model_artifact_feature_stats is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Cannot perform drift analysis as there is no sample set to compare to. The model artifact was not "
                "logged with a sample set and `sample_set` was not provided to the function."
            )
        # Return the statistics logged with the model:
        return model_artifact_feature_stats

    # Turn other object types to DataFrame:
    # A DataFrame is necessary for executing the "drop features" operation.
    dataset_types = list(DatasetType.__args__)
    if typing.Any in dataset_types:
        dataset_types.remove(typing.Any)
    if isinstance(
        sample_set,
        tuple(dataset_types),
    ):
        sample_set, _ = read_dataset_as_dataframe(
            dataset=sample_set,
            feature_columns=sample_set_columns,
            drop_columns=sample_set_drop_columns,
            label_columns=sample_set_label_columns,
        )
    else:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Parameter sample_set has an unsupported type: {type(sample_set)}"
        )

    # Return the sample set statistics:
    return get_df_stats(df=sample_set, options=InferOptions.Histogram)


def read_dataset_as_dataframe(
    dataset: DatasetType,
    feature_columns: typing.Optional[typing.Union[str, list[str]]] = None,
    label_columns: typing.Optional[typing.Union[str, list[str]]] = None,
    drop_columns: typing.Optional[typing.Union[str, list[str], int, list[int]]] = None,
) -> tuple[pd.DataFrame, list[str]]:
    """
    Parse the given dataset into a DataFrame and drop the columns accordingly. In addition, the label columns will be
    parsed and validated as well.

    :param dataset:         A dataset that will be converted into a DataFrame.
                            Can be either a list of lists, numpy.ndarray, dict, pd.Series, DataItem
                            or a FeatureVector.
    :param feature_columns: List of feature columns that will be used to build the dataframe when dataset is from
                            type list or numpy array.
    :param label_columns:   The target label(s) of the column(s) in the dataset. for Regression or
                            Classification tasks.
    :param drop_columns:    ``str`` / ``int`` or a list of ``str`` / ``int`` that represent the column names / indices
                            to drop.

    :returns: A tuple of:
              [0] = The parsed dataset as a DataFrame
              [1] = Label columns.

    raises MLRunInvalidArgumentError: If the `drop_columns` are not matching the dataset or unsupported dataset type.
    """
    # Turn the `drop labels` into a list if given:
    if drop_columns is not None:
        if not isinstance(drop_columns, list):
            drop_columns = [drop_columns]

    # Check if the dataset is in fact a Feature Vector:
    if isinstance(dataset, mlrun.feature_store.FeatureVector):
        # Try to get the label columns if not provided:
        if label_columns is None:
            label_columns = dataset.status.label_column
        # Get the features and parse to DataFrame:
        dataset = dataset.get_offline_features(drop_columns=drop_columns).to_dataframe()

    elif isinstance(dataset, (list, np.ndarray)):
        if not feature_columns:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Feature columns list must be provided when dataset input as from type list or numpy array"
            )
        # Parse the list / numpy array into a DataFrame:
        dataset = pd.DataFrame(dataset, columns=feature_columns)
        # Validate the `drop_columns` is given as integers:
        if drop_columns and not all(isinstance(col, int) for col in drop_columns):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`drop_columns` must be an integer / list of integers if provided as a list."
            )
    elif isinstance(dataset, mlrun.DataItem):
        if (
            not dataset.url
            and dataset.artifact_url
            and mlrun.datastore.parse_store_uri(dataset.artifact_url)[0]
            == mlrun.utils.StorePrefix.FeatureVector
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"No data has been found. Make sure you have applied `get_offline_features` "
                f"on your feature vector {dataset.artifact_url} with a valid target before passing "
                f"it as an input."
            )
        # Turn the DataItem to DataFrame:
        dataset = dataset.as_df()
    else:
        # Parse the object (should be a pd.DataFrame / pd.Series, dictionary) into a DataFrame:
        try:
            dataset = pd.DataFrame(dataset)
        except ValueError as e:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Could not parse the given dataset of type {type(dataset)} into a pandas DataFrame. "
                f"Received the following error: {e}"
            )
    # Drop columns if needed:
    if drop_columns:
        dataset.drop(drop_columns, axis=1, inplace=True)

    # Turn the `label_columns` into a list by default:
    if label_columns is None:
        label_columns = []
    elif isinstance(label_columns, (str, int)):
        label_columns = [label_columns]

    return dataset, label_columns


def log_result(
    context: "mlrun.MLClientCtx",
    result_set_name: str,
    result_set: pd.DataFrame,
    artifacts_tag: str,
    batch_id: str,
) -> None:
    # Log the result set:
    context.logger.info("Logging result set (x | prediction)...")
    context.log_dataset(
        key=result_set_name,
        df=result_set,
        db_key=result_set_name,
        tag=artifacts_tag,
    )
    # Log the batch ID:
    if batch_id is None:
        batch_id = hashlib.sha224(str(datetime_now()).encode()).hexdigest()
    context.log_result(
        key="batch_id",
        value=batch_id,
    )


def _create_model_monitoring_function_base(
    *,
    project: str,
    func: typing.Union[str, None] = None,
    application_class: typing.Union[
        str, "mm_app.ModelMonitoringApplicationBase", None
    ] = None,
    name: typing.Optional[str] = None,
    image: typing.Optional[str] = None,
    tag: typing.Optional[str] = None,
    requirements: typing.Union[list[str], None] = None,
    requirements_file: str = "",
    local_path: typing.Optional[str] = None,
    **application_kwargs,
) -> mlrun.runtimes.ServingRuntime:
    """
    Note: this is an internal API only.
    This function does not set the labels or mounts v3io.
    """
    if name in mm_constants._RESERVED_FUNCTION_NAMES:
        raise mlrun.errors.MLRunValueError(
            "An application cannot have the following names: "
            f"{mm_constants._RESERVED_FUNCTION_NAMES}"
        )
    _, has_valid_suffix, suffix = mlrun.utils.helpers.ensure_batch_job_suffix(name)
    if name and not has_valid_suffix:
        raise mlrun.errors.MLRunValueError(
            f"Model monitoring application names cannot end with `{suffix}`"
        )
    if func is None:
        func = ""
    if check_if_hub_uri(func):
        hub_module = mlrun.get_hub_module(url=func, local_path=local_path)
        if hub_module.kind != HubModuleType.monitoring_app:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The provided module is not a monitoring application"
            )
        requirements = mlrun.model.ImageBuilder.resolve_requirements(
            requirements, requirements_file
        )
        requirements = merge_requirements(
            reqs_priority=requirements, reqs_secondary=hub_module.requirements
        )
        func = hub_module.get_module_file_path()
    func_obj = typing.cast(
        mlrun.runtimes.ServingRuntime,
        mlrun.code_to_function(
            filename=func,
            name=name,
            project=project,
            tag=tag,
            kind=mlrun.run.RuntimeKinds.serving,
            image=image,
            requirements=requirements,
            requirements_file=requirements_file,
        ),
    )
    graph = func_obj.set_topology(mlrun.serving.states.StepKinds.flow)
    prepare_step = graph.to(
        class_name="mlrun.model_monitoring.applications._application_steps._PrepareMonitoringEvent",
        name="PrepareMonitoringEvent",
        application_name=name,
    )
    if isinstance(application_class, str):
        app_step = prepare_step.to(class_name=application_class, **application_kwargs)
    else:
        app_step = prepare_step.to(class_name=application_class)

    app_step.__class__ = mlrun.serving.MonitoringApplicationStep

    app_step.error_handler(
        class_name="mlrun.model_monitoring.applications._application_steps._ApplicationErrorHandler",
        name="ApplicationErrorHandler",
        full_event=True,
        project=project,
    )

    app_step.to(
        class_name="mlrun.model_monitoring.applications._application_steps._PushToMonitoringWriter",
        name="PushToMonitoringWriter",
        project=project,
    )

    def block_to_mock_server(*args, **kwargs) -> typing.NoReturn:
        raise NotImplementedError(
            "Model monitoring serving functions do not support `.to_mock_server`. "
            "You may call your model monitoring application object logic via the `.evaluate` method."
        )

    func_obj.to_mock_server = block_to_mock_server  # Until ML-7643 is implemented

    return func_obj



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import typing
from io import BytesIO

import mlrun

from .base import Artifact

if typing.TYPE_CHECKING:
    from plotly.graph_objs import Figure


class PlotArtifact(Artifact):
    kind = "plot"

    _TEMPLATE = """
<h3 style="text-align:center">{}</h3>
<img title="{}" src="data:image/png;base64,{}">
"""

    def __init__(
        self, key=None, body=None, is_inline=False, target_path=None, title=None
    ):
        super().__init__(key, body, format="html", target_path=target_path)
        self.metadata.description = title

    def before_log(self):
        self.spec.viewer = "chart"
        import matplotlib

        if not self.spec.get_body() or not isinstance(
            self.spec.get_body(), (bytes, matplotlib.figure.Figure)
        ):
            raise ValueError(
                "matplotlib fig or png bytes must be provided as artifact body"
            )

    def get_body(self):
        """Convert Matplotlib figure 'fig' into a <img> tag for HTML use
        using base64 encoding."""
        if isinstance(self.spec.get_body(), bytes):
            data = self.spec.get_body()
        else:
            from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

            canvas = FigureCanvas(self.spec.get_body())
            png_output = BytesIO()
            canvas.print_png(png_output)
            data = png_output.getvalue()

        data_uri = base64.b64encode(data).decode("utf-8")
        return self._TEMPLATE.format(
            self.metadata.description or self.metadata.key, self.metadata.key, data_uri
        )


class PlotlyArtifact(Artifact):
    """
    Plotly artifact is an artifact for saving Plotly generated figures. They will be stored in a html format.
    """

    kind = "plotly"

    def __init__(
        self,
        figure: typing.Optional["Figure"] = None,
        key: typing.Optional[str] = None,
        target_path: typing.Optional[str] = None,
    ) -> None:
        """
        Initialize a Plotly artifact with the given figure.

        :param figure:      Plotly figure ('plotly.graph_objs.Figure' object) to save as an artifact.
        :param key:         Key for the artifact to be stored in the database.
        :param target_path: Path to save the artifact.
        """
        # Validate the plotly package:
        try:
            from plotly.graph_objs import Figure
        except ModuleNotFoundError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "Using `PlotlyArtifact` requires plotly package. Use `pip install mlrun[plotly]` to install it."
            )
        except ImportError:
            import plotly

            raise mlrun.errors.MLRunMissingDependencyError(
                f"Using `PlotlyArtifact` requires plotly version >= 5.4.0 but found version {plotly.__version__}. "
                f"Use `pip install -U mlrun[plotly]` to install it."
            )

        # Call the artifact initializer:
        super().__init__(key=key, target_path=target_path, viewer="plotly")

        # Validate input:
        if figure is not None and not isinstance(figure, Figure):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"PlotlyArtifact requires the figure parameter to be a "
                f"`plotly.graph_objs.Figure` but received '{type(figure)}'"
            )

        # Continue initializing the plotly artifact:
        self._figure = figure
        self.spec.format = "html"

    def get_body(self) -> str:
        """
        Get the artifact's body - the Plotly figure's html code.

        :return: The figure's html code.
        """
        return self._figure.to_html()



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import Optional, Union

import mlrun
import mlrun.common.schemas.alert as alert_objects
from mlrun.model import ModelObj


class AlertConfig(ModelObj):
    _dict_fields = [
        "project",
        "name",
        "description",
        "summary",
        "severity",
        "reset_policy",
        "state",
        "count",
        "created",
        "updated",
    ]
    _fields_to_serialize = ModelObj._fields_to_serialize + [
        "entities",
        "notifications",
        "trigger",
        "criteria",
    ]

    def __init__(
        self,
        project: Optional[str] = None,
        name: Optional[str] = None,
        template: Union[alert_objects.AlertTemplate, str] = None,
        description: Optional[str] = None,
        summary: Optional[str] = None,
        severity: alert_objects.AlertSeverity = None,
        trigger: alert_objects.AlertTrigger = None,
        criteria: alert_objects.AlertCriteria = None,
        reset_policy: alert_objects.ResetPolicy = None,
        notifications: Optional[list[alert_objects.AlertNotification]] = None,
        entities: alert_objects.EventEntities = None,
        id: Optional[int] = None,
        state: alert_objects.AlertActiveState = None,
        created: Optional[str] = None,
        count: Optional[int] = None,
        updated: Optional[str] = None,
        **kwargs,
    ):
        """Alert config object

        Example::

            # create an alert on endpoint_id, which will be triggered to slack if there is a "data_drift_detected" event
            # 3 times in the next hour.

            from mlrun.alerts import AlertConfig
            import mlrun.common.schemas.alert as alert_objects

            entity_kind = alert_objects.EventEntityKind.MODEL_ENDPOINT_RESULT
            entity_id = get_default_result_instance_fqn(endpoint_id)
            event_name = alert_objects.EventKind.DATA_DRIFT_DETECTED
            notification = mlrun.model.Notification(
                kind="slack",
                name="slack_notification",
                message="drift was detected",
                severity="warning",
                when=["now"],
                condition="failed",
                secret_params={
                    "webhook": "https://hooks.slack.com/",
                },
            ).to_dict()

            alert_data = AlertConfig(
                project="my-project",
                name="drift-alert",
                summary="a drift was detected",
                severity=alert_objects.AlertSeverity.LOW,
                entities=alert_objects.EventEntities(
                    kind=entity_kind, project="my-project", ids=[entity_id]
                ),
                trigger=alert_objects.AlertTrigger(events=[event_name]),
                criteria=alert_objects.AlertCriteria(count=3, period="1h"),
                notifications=[alert_objects.AlertNotification(notification=notification)],
            )
            project.store_alert_config(alert_data)

        :param project:        Name of the project to associate the alert with
        :param name:           Name of the alert
        :param template:       Optional parameter that allows creating an alert based on a predefined template.
                               You can pass either an AlertTemplate object or a string (the template name).
                               If a template is used, many fields of the alert will be auto-generated based on the
                               template.However, you still need to provide the following fields:
                               `name`, `project`, `entity`, `notifications`
        :param description:    Description of the alert
        :param summary:        Summary of the alert, will be sent in the generated notifications
        :param severity:       Severity of the alert
        :param trigger:        The events that will trigger this alert, may be a simple trigger based on events or
                               complex trigger which is based on a prometheus alert
        :param criteria:       When the alert will be triggered based on the specified number of events within the
                               defined time period.
        :param reset_policy:   When to clear the alert. Either "manual" for manual reset of the alert, or
                               "auto" if the criteria contains a time period
        :param notifications:  List of notifications to invoke once the alert is triggered
        :param entities:       Entities that the event relates to. The entity object will contain fields that
                               uniquely identify a given entity in the system
        :param id:             Internal id of the alert (user should not supply it)
        :param state:          State of the alert, may be active/inactive (user should not supply it)
        :param created:        When the alert is created (user should not supply it)
        :param count:          Internal counter of the alert (user should not supply it)
        :param updated:        The last update time of the alert (user should not supply it)
        """
        self.project = project
        self.name = name
        self.description = description
        self.summary = summary
        self.severity = severity
        self.trigger = trigger
        self.criteria = criteria
        self.reset_policy = reset_policy
        self.notifications = notifications or []
        self.entities = entities
        self.id = id
        self.state = state
        self._created = created
        self.count = count
        self._updated = updated

        if template:
            self._apply_template(template)

    @property
    def created(self) -> datetime:
        """
        Get the `created` field as a datetime object.
        """
        if isinstance(self._created, str):
            return datetime.fromisoformat(self._created)
        return self._created

    @created.setter
    def created(self, created):
        self._created = created

    @property
    def updated(self) -> datetime:
        """
        Get the `updated` field as a datetime object.
        """
        if isinstance(self._updated, str):
            return datetime.fromisoformat(self._updated)
        return self._updated

    @updated.setter
    def updated(self, updated):
        self._updated = updated

    def validate_required_fields(self):
        if not self.name:
            raise mlrun.errors.MLRunInvalidArgumentError("Alert name must be provided")

    def _serialize_field(
        self, struct: dict, field_name: Optional[str] = None, strip: bool = False
    ):
        if field_name == "entities":
            if self.entities:
                return (
                    self.entities.dict()
                    if not isinstance(self.entities, dict)
                    else self.entities
                )
            return None
        if field_name == "notifications":
            if self.notifications:
                return [
                    notification_data.dict()
                    if not isinstance(notification_data, dict)
                    else notification_data
                    for notification_data in self.notifications
                ]
            return None
        if field_name == "trigger":
            if self.trigger:
                return (
                    self.trigger.dict()
                    if not isinstance(self.trigger, dict)
                    else self.trigger
                )
            return None
        if field_name == "criteria":
            if self.criteria:
                return (
                    self.criteria.dict()
                    if not isinstance(self.criteria, dict)
                    else self.criteria
                )
            return None
        return super()._serialize_field(struct, field_name, strip)

    def to_dict(
        self,
        fields: Optional[list] = None,
        exclude: Optional[list] = None,
        strip: bool = False,
    ):
        if self.entities is None:
            raise mlrun.errors.MLRunBadRequestError("Alert entity field is missing")
        if not self.notifications:
            raise mlrun.errors.MLRunBadRequestError(
                "Alert must have at least one notification"
            )
        return super().to_dict(self._dict_fields)

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: Optional[dict] = None
    ):
        new_obj = super().from_dict(struct, fields=fields)

        entity_data = struct.get("entities")
        if entity_data:
            entity_obj = alert_objects.EventEntities.parse_obj(entity_data)
            new_obj.entities = entity_obj

        notifications_data = struct.get("notifications")
        if notifications_data:
            notifications_objs = [
                alert_objects.AlertNotification.parse_obj(notification)
                for notification in notifications_data
            ]
            new_obj.notifications = notifications_objs

        trigger_data = struct.get("trigger")
        if trigger_data:
            trigger_obj = alert_objects.AlertTrigger.parse_obj(trigger_data)
            new_obj.trigger = trigger_obj

        criteria_data = struct.get("criteria")
        if criteria_data:
            criteria_obj = alert_objects.AlertCriteria.parse_obj(criteria_data)
            new_obj.criteria = criteria_obj
        return new_obj

    def with_notifications(self, notifications: list[alert_objects.AlertNotification]):
        if not isinstance(notifications, list) or not all(
            isinstance(item, alert_objects.AlertNotification) for item in notifications
        ):
            raise ValueError(
                "Notifications parameter must be a list of AlertNotification"
            )
        for notification_data in notifications:
            self.notifications.append(notification_data)
        return self

    def with_entities(self, entities: alert_objects.EventEntities):
        if not isinstance(entities, alert_objects.EventEntities):
            raise ValueError("Entities parameter must be of type: EventEntities")
        self.entities = entities
        return self

    def _apply_template(self, template):
        if isinstance(template, str):
            db = mlrun.get_run_db()
            template = db.get_alert_template(template)

        # Apply parameters from the template to the AlertConfig object only if they are not already specified by the
        # user in the current configuration.
        # User-provided parameters will take precedence over corresponding template values
        self.summary = self.summary or template.summary
        self.severity = self.severity or template.severity
        self.criteria = self.criteria or template.criteria
        self.trigger = self.trigger or template.trigger
        self.reset_policy = self.reset_policy or template.reset_policy

    def list_activations(
        self,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None,
        from_last_update: bool = False,
    ) -> list[mlrun.common.schemas.alert.AlertActivation]:
        """
        Retrieve a list of all alert activations.

        :param since: Filters for alert activations occurring after this timestamp.
        :param until: Filters for alert activations occurring before this timestamp.
        :param from_last_update: If set to True, retrieves alert activations since the alert's last update time.
                                 if both since and from_last_update=True are provided, from_last_update takes precedence
                                 and the since value will be overridden by the alert's last update timestamp.

        :returns: A list of alert activations matching the provided filters.
        """
        db = mlrun.get_run_db()
        if from_last_update and self._updated:
            since = self.updated

        return db.list_alert_activations(
            project=self.project,
            name=self.name,
            since=since,
            until=until,
        )

    def paginated_list_activations(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        from_last_update: bool = False,
        **kwargs,
    ) -> tuple[mlrun.common.schemas.alert.AlertActivation, Optional[str]]:
        """
        List alerts activations with support for pagination and various filtering options.

        This method retrieves a paginated list of alert activations based on the specified filter parameters.
        Pagination is controlled using the `page`, `page_size`, and `page_token` parameters. The method
        will return a list of alert activations that match the filtering criteria provided.

        For detailed information about the parameters, refer to the list_activations method:
            See :py:func:`~list_activations` for more details.

        Examples::

            # Fetch first page of alert activations with page size of 5
            alert_activations, token = alert_config.paginated_list_activations(page_size=5)
            # Fetch next page using the pagination token from the previous response
            alert_activations, token = alert_config.paginated_list_activations(
                page_token=token
            )
            # Fetch alert activations for a specific page (e.g., page 3)
            alert_activations, token = alert_config.paginated_list_activations(
                page=3, page_size=5
            )

            # Automatically iterate over all pages without explicitly specifying the page number
            alert_activations = []
            token = None
            while True:
                page_alert_activations, token = alert_config.paginated_list_activations(
                    page_token=token, page_size=5
                )
                alert_activations.extend(page_alert_activations)

                # If token is None and page_alert_activations is empty, we've reached the end (no more activations).
                # If token is None and page_alert_activations is not empty, we've fetched the last page of activations.
                if not token:
                    break
            print(f"Total alert activations retrieved: {len(alert_activations)}")

        :param page: The page number to retrieve. If not provided, the next page will be retrieved.
        :param page_size: The number of items per page to retrieve. Up to `page_size` responses are expected.
        :param page_token: A pagination token used to retrieve the next page of results. Should not be provided
            for the first request.
        :param from_last_update: If set to True, retrieves alert activations since the alert's last update time.

        :returns: A tuple containing the list of alert activations and an optional `page_token` for pagination.
        """
        if from_last_update and self._updated:
            kwargs["since"] = self.updated

        db = mlrun.get_run_db()
        return db.paginated_list_alert_activations(
            *args,
            project=self.project,
            name=self.name,
            page=page,
            page_size=page_size,
            page_token=page_token,
            **kwargs,
        )



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys
from pathlib import Path
from typing import Optional, Union

import yaml

import mlrun.common.types
import mlrun.utils
from mlrun.common.schemas.hub import HubModuleType, HubSourceType
from mlrun.run import function_to_module, get_object
from mlrun.utils import logger

from ..errors import MLRunBadRequestError
from ..model import ModelObj
from ..utils import extend_hub_uri_if_needed


class HubModule(ModelObj):
    def __init__(
        self,
        name: str,
        kind: Union[HubModuleType, str],
        version: Optional[str] = None,
        description: Optional[str] = None,
        categories: Optional[list] = None,
        requirements: Optional[list] = None,
        local_path: Optional[str] = None,
        filename: Optional[str] = None,
        example: Optional[str] = None,
        url: Optional[str] = None,
        **kwargs,  # catch all for unused args
    ):
        self.name: str = name
        self.version: str = version
        self.kind: HubModuleType = kind
        self.description: str = description or ""
        self.categories: list = categories or []
        self.requirements: list = requirements or []
        self.local_path: str = local_path or ""
        self.filename: str = filename or name + ".py"
        self.example: str = example or ""
        self.url: str = url or ""

    def module(self):
        """Import the module after downloading its fils to local_path"""
        try:
            return function_to_module(code=self.filename, workdir=self.local_path)
        except FileNotFoundError:
            searched_path = self.local_path or "./"
            raise FileNotFoundError(
                f"Module file {self.filename} not found in {searched_path}, try calling download_module_files() first"
            )

    def install_requirements(self) -> None:
        """
        Install pip-style requirements (e.g., ["pandas>=2.0.0", "requests==2.31.0"]).
        """
        for req in self.requirements:
            logger.info(f"Installing {req} ...")
            try:
                subprocess.run(
                    [sys.executable, "-m", "pip", "install", req], check=True, text=True
                )
                logger.info(f"Installed {req}")
            except subprocess.CalledProcessError as e:
                logger.error(f"Failed to install {req} (exit code {e.returncode})")

    def download_module_files(self, local_path=None, secrets=None):
        """
        Download this hub module’s files (code file and, if available, an example notebook) to the target directory
        specified by `local_path` (defaults to the current working directory).
        This path will be used later to locate the code file when importing the module.
        """
        self.local_path = self.verify_directory(path=local_path)
        source_url, _ = extend_hub_uri_if_needed(
            uri=self.url, asset_type=HubSourceType.modules, file=self.filename
        )
        self._download_object(
            obj_url=source_url, target_name=self.filename, secrets=secrets
        )
        if self.example:
            example_url, _ = extend_hub_uri_if_needed(
                uri=self.url, asset_type=HubSourceType.modules, file=self.example
            )
            self._download_object(
                obj_url=example_url, target_name=self.example, secrets=secrets
            )

    def _download_object(self, obj_url, target_name, secrets=None):
        data = get_object(url=obj_url, secrets=secrets)
        target_dir = self.local_path if self.local_path is not None else os.getcwd()
        target_filepath = os.path.join(target_dir, target_name)
        with open(target_filepath, "wb") as f:
            f.write(data)

    @staticmethod
    def verify_directory(path: Optional[str] = None) -> Path:
        """
        Validate that the given path is an existing directory.
        If no path has been provided, returns current working directory.
        """
        if path:
            path = Path(path)
            if not path.exists():
                raise ValueError(f"Path does not exist: {path}")
            if not path.is_dir():
                raise ValueError(f"Path is not a directory: {path}")
            return path
        return Path(os.getcwd())

    def get_module_file_path(self):
        if not self.local_path:
            raise MLRunBadRequestError(
                "module files haven't been downloaded yet, try calling download_module_files() first"
            )
        return str(Path(self.local_path) / self.filename)


def get_hub_module(
    url: str = "",
    download_files: Optional[bool] = True,
    secrets: Optional[dict] = None,
    local_path: Optional[str] = None,
) -> HubModule:
    """
    Get a hub-module object containing metadata of the requested module.
    :param url: Hub module url in the format "hub://[<source>/]<item-name>[:<tag>]"
    :param download_files: When set to True, the module files (code file and example notebook) are downloaded
    :param secrets: Optional, credentials dict for DB or URL (s3, v3io, ...)
    :param local_path: Path to target directory for the module files. Ignored when download_files is set to False.
                       Defaults to the current working directory.

    :return: HubModule object
    """
    item_yaml_url, is_hub_uri = extend_hub_uri_if_needed(
        uri=url, asset_type=HubSourceType.modules, file="item.yaml"
    )
    if not is_hub_uri:
        raise mlrun.errors.MLRunInvalidArgumentError("Not a valid hub URL")
    yaml_obj = get_object(url=item_yaml_url, secrets=secrets)
    item_yaml = yaml.safe_load(yaml_obj)
    spec = item_yaml.pop("spec", {})
    hub_module = HubModule(**item_yaml, **spec, url=url)
    if download_files:
        hub_module.download_module_files(local_path=local_path, secrets=secrets)
    return hub_module


def import_module(url="", install_requirements=False, secrets=None, local_path=None):
    """
    Import a module from the hub to use directly.
    :param url: hub module url in the format "hub://[<source>/]<item-name>[:<tag>]"
    :param install_requirements: when set to True, the module's requirements are installed.
    :param secrets: optional, credentials dict for DB or URL (s3, v3io, ...)
    :param local_path: Path to target directory for the module files (code and example notebook).
                       Defaults to the current working directory.

    :return: the module
    """
    hub_module: HubModule = get_hub_module(
        url=url, download_files=True, secrets=secrets, local_path=local_path
    )
    if install_requirements:
        hub_module.install_requirements()
    return hub_module.module()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from os import listdir, makedirs, path, stat
from shutil import copyfile
from typing import Optional

import fsspec

import mlrun

from .base import DataStore, FileStats


class FileStore(DataStore):
    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, "file", endpoint, secrets=secrets)

        self._item_path, self._real_path = None, None
        if mlrun.mlconf.storage.item_to_real_path:
            # map item path prefix with real (local) path prefix e.g.:
            # item_to_real_path="\data::c:\\mlrun_data" replace \data\x.csv with c:\\mlrun_data\x.csv
            split = mlrun.mlconf.storage.item_to_real_path.split("::")
            self._item_path = split[0].strip()
            self._real_path = split[1].strip().rstrip("/").rstrip("\\")

    @property
    def url(self):
        return self.subpath

    def _join(self, key: str):
        if self._item_path and not self.subpath:
            if key.startswith(self._item_path):
                suffix = key[len(self._item_path) :]
                if suffix[0] in ["/", "\\"]:
                    suffix = suffix[1:]
                key = path.join(self._real_path, suffix)
        return path.join(self.subpath, key)

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if not self._filesystem:
            self._filesystem = fsspec.filesystem("file")
        return self._filesystem

    def get(self, key, size=None, offset=0):
        with open(self._join(key), "rb") as fp:
            if offset:
                fp.seek(offset)
            if not size:
                size = -1
            return fp.read(size)

    def put(self, key, data, append=False):
        dir_to_create = path.dirname(self._join(key))
        if dir_to_create:
            self._ensure_directory(dir_to_create)
        data, mode = self._prepare_put_data(data, append)
        with open(self._join(key), mode) as fp:
            fp.write(data)
            fp.close()

    def download(self, key, target_path):
        fullpath = self._join(key)
        if fullpath == target_path:
            return
        copyfile(fullpath, target_path)

    def upload(self, key, src_path):
        fullpath = self._join(key)
        if path.realpath(src_path) == path.realpath(fullpath):
            return
        dir = path.dirname(fullpath)
        if dir:
            makedirs(dir, exist_ok=True)
        copyfile(src_path, fullpath)

    def stat(self, key):
        s = stat(self._join(key))
        return FileStats(size=s.st_size, modified=s.st_mtime)

    def listdir(self, key):
        return listdir(self._join(key))

    def _ensure_directory(self, dir_to_create):
        # We retry the makedirs because it can fail if another process is creating the same dir
        # Note - inside it try to catch FileExistsError, but it still fails sometimes during its internal logic
        # where it calls `mkdir()`.
        while True:
            try:
                makedirs(dir_to_create, exist_ok=True)
                return
            except FileExistsError:
                time.sleep(0.1)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import mlrun

from .base import BaseMerger


class DaskFeatureMerger(BaseMerger):
    engine = "dask"
    support_offline = True

    def __init__(self, vector, **engine_args):
        super().__init__(vector, **engine_args)
        try:
            import dask  # noqa: F401
        except (ModuleNotFoundError, ImportError) as exc:
            raise ImportError(
                "Using 'DaskFeatureMerger' requires dask package. Use pip install mlrun[dask] to install it."
            ) from exc

        self.client = engine_args.get("dask_client")
        self._dask_cluster_uri = engine_args.get("dask_cluster_uri")

    def _reset_index(self, df):
        to_drop = df.index.name is None
        df = df.reset_index(drop=to_drop)
        return df

    def _asof_join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name: str,
        featureset_timestamp: str,
        featureset_df: list,
        left_keys: list,
        right_keys: list,
    ):
        from dask.dataframe.multi import merge_asof

        featureset_df = self._normalize_timestamp_column(
            entity_timestamp_column,
            entity_df,
            featureset_timestamp,
            featureset_df,
            featureset_name,
        )

        def sort_partition(partition, timestamp):
            return partition.sort_values(timestamp)

        entity_df = entity_df.map_partitions(
            sort_partition, timestamp=entity_timestamp_column
        )
        featureset_df = featureset_df.map_partitions(
            sort_partition, timestamp=featureset_timestamp
        )

        merged_df = merge_asof(
            entity_df,
            featureset_df,
            left_on=entity_timestamp_column,
            right_on=featureset_timestamp,
            left_by=left_keys or None,
            right_by=right_keys or None,
            suffixes=("", f"_{featureset_name}_"),
        )
        for col in merged_df.columns:
            if re.findall(f"_{featureset_name}_$", col):
                self._append_drop_column(col)

        return merged_df

    def _join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name,
        featureset_timestamp,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        from dask.dataframe.multi import merge

        merged_df = merge(
            entity_df,
            featureset_df,
            how=self._join_type,
            left_on=left_keys,
            right_on=right_keys,
            suffixes=("", f"_{featureset_name}_"),
        )
        for col in merged_df.columns:
            if re.findall(f"_{featureset_name}_$", col):
                self._append_drop_column(col)
        return merged_df

    def get_status(self):
        if self._result_df is None:
            raise RuntimeError("unexpected status, no result df")
        return "completed"

    def get_df(self, to_pandas=True):
        if to_pandas and hasattr(self._result_df, "dask"):
            df = self._result_df.compute()
        else:
            df = self._result_df
        self._set_indexes(df)
        return df

    def _create_engine_env(self):
        from dask.distributed import Client

        if "index" not in self._index_columns:
            self._append_drop_column("index")

        # init the dask client if needed
        if not self.client:
            if self._dask_cluster_uri:
                function = mlrun.import_function(self._dask_cluster_uri)
                self.client = function.client
            else:
                self.client = Client()

    def _get_engine_df(
        self,
        feature_set,
        feature_set_name,
        column_names=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
    ):
        import dask.dataframe as dd

        df = feature_set.to_dataframe(
            columns=column_names,
            df_module=dd,
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            index=False,
            additional_filters=additional_filters,
        )

        return self._reset_index(df).persist()

    def _rename_columns_and_select(self, df, rename_col_dict, columns=None):
        return df.rename(
            columns=rename_col_dict,
        )

    def _drop_columns_from_result(self):
        self._result_df = self._result_df.drop(
            columns=self._drop_columns, errors="ignore"
        )

    def _filter(self, query):
        self._result_df = self._result_df.query(query)

    def _order_by(self, order_by_active):
        self._result_df.sort_values(by=order_by_active)

    def _convert_entity_rows_to_engine_df(self, entity_rows):
        import dask.dataframe as dd

        if entity_rows is not None and not hasattr(entity_rows, "dask"):
            return dd.from_pandas(entity_rows, npartitions=len(entity_rows.columns))

        return entity_rows



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
from copy import copy

import mlrun
import mlrun.errors
from mlrun.common.schemas import AuthorizationVerificationInput
from mlrun.runtimes import BaseRuntime
from mlrun.runtimes.function_reference import FunctionReference
from mlrun.runtimes.utils import enrich_function_from_dict
from mlrun.utils import StorePrefix, logger

from ..common.helpers import parse_versioned_object_uri
from ..config import config

project_separator = "/"
feature_separator = "."
expected_message = f"in the form feature-set{feature_separator}feature[ as alias]"


def parse_feature_string(feature):
    """parse feature string into feature set name, feature name, alias"""
    # expected format: <feature-set>.<name|*>[ as alias]
    if feature_separator not in feature:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"feature {feature} must be {expected_message}"
        )
    feature_set, feature_name = feature.rsplit(feature_separator, 1)
    feature_set = feature_set.strip()
    split_result = feature_name.split(" as ", 1)
    feature_name = split_result[0].strip()
    alias = split_result[1].strip() if len(split_result) > 1 else None
    return feature_set, feature_name, alias


def parse_project_name_from_feature_string(feature):
    """parse feature string into project name and feature"""
    # expected format: <project-name>/<feature>
    if project_separator not in feature:
        return None, feature

    splitted = feature.split(project_separator)
    if len(splitted) > 2:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"feature {feature} must be {expected_message}, cannot have more than one '/'"
        )
    project_name = splitted[0]
    feature_name = splitted[1]
    return project_name.strip(), feature_name.strip()


def parse_feature_set_uri(uri, project=None):
    """get feature set object from db by uri"""
    active_project = project or config.active_project

    # parse store://.. uri
    if mlrun.datastore.is_store_uri(uri):
        prefix, new_uri = mlrun.datastore.parse_store_uri(uri)
        if prefix != StorePrefix.FeatureSet:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"provided store uri ({uri}) does not represent a feature set (prefix={prefix})"
            )
        uri = new_uri

    return parse_versioned_object_uri(uri, active_project)


def get_feature_set_by_uri(uri, project=None):
    """get feature set object from db by uri"""
    db = mlrun.get_run_db()
    project, name, tag, uid = parse_feature_set_uri(uri, project)
    resource = (
        mlrun.common.schemas.AuthorizationResourceTypes.feature_set.to_resource_string(
            project, "feature-set"
        )
    )

    auth_input = AuthorizationVerificationInput(
        resource=resource, action=mlrun.common.schemas.AuthorizationAction.read
    )
    db.verify_authorization(auth_input)

    return db.get_feature_set(name, project, tag, uid)


def get_feature_vector_by_uri(uri, project=None, update=True):
    """get feature vector object from db by uri"""
    db = mlrun.get_run_db()
    active_project = project or config.active_project

    # parse store://.. uri
    if mlrun.datastore.is_store_uri(uri):
        prefix, new_uri = mlrun.datastore.parse_store_uri(uri)
        if prefix != StorePrefix.FeatureVector:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"provided store uri ({uri}) does not represent a feature vector (prefix={prefix})"
            )
        uri = new_uri

    project, name, tag, uid = parse_versioned_object_uri(uri, active_project)

    resource = mlrun.common.schemas.AuthorizationResourceTypes.feature_vector.to_resource_string(
        project, "feature-vector"
    )

    if update:
        auth_input = AuthorizationVerificationInput(
            resource=resource, action=mlrun.common.schemas.AuthorizationAction.update
        )
    else:
        auth_input = AuthorizationVerificationInput(
            resource=resource, action=mlrun.common.schemas.AuthorizationAction.read
        )

    db.verify_authorization(auth_input)

    return db.get_feature_vector(name, project, tag, uid)


def verify_feature_set_permissions(
    feature_set, action: mlrun.common.schemas.AuthorizationAction
):
    project, _, _, _ = parse_feature_set_uri(feature_set.uri)

    resource = (
        mlrun.common.schemas.AuthorizationResourceTypes.feature_set.to_resource_string(
            project, "feature-set"
        )
    )
    db = feature_set._get_run_db()

    auth_input = AuthorizationVerificationInput(resource=resource, action=action)
    db.verify_authorization(auth_input)


def verify_feature_set_exists(feature_set):
    db = feature_set._get_run_db()
    project, uri, tag, _ = parse_feature_set_uri(feature_set.uri)

    try:
        fset = db.get_feature_set(feature_set.metadata.name, project, tag)
    except mlrun.errors.MLRunNotFoundError:
        raise mlrun.errors.MLRunNotFoundError(f"feature set {uri} does not exist")

    if not fset.spec.features:
        raise mlrun.errors.MLRunNotFoundError(f"feature set {uri} is empty")


def verify_feature_vector_permissions(
    feature_vector, action: mlrun.common.schemas.AuthorizationAction
):
    project = feature_vector._metadata.project or config.active_project

    resource = mlrun.common.schemas.AuthorizationResourceTypes.feature_vector.to_resource_string(
        project, "feature-vector"
    )

    db = mlrun.get_run_db()
    auth_input = AuthorizationVerificationInput(resource=resource, action=action)
    db.verify_authorization(auth_input)


class RunConfig:
    """class for holding function and run specs for jobs and serving functions"""

    def __init__(
        self,
        function: typing.Union[str, FunctionReference, BaseRuntime] = None,
        local: typing.Optional[bool] = None,
        image: typing.Optional[str] = None,
        kind: typing.Optional[str] = None,
        handler: typing.Optional[str] = None,
        parameters: typing.Optional[dict] = None,
        watch: typing.Optional[bool] = None,
        owner=None,
        credentials: typing.Optional[mlrun.model.Credentials] = None,
        code: typing.Optional[str] = None,
        requirements: typing.Optional[typing.Union[str, list[str]]] = None,
        extra_spec: typing.Optional[dict] = None,
        auth_info=None,
    ):
        """class for holding function and run specs for jobs and serving functions

        when running feature ingestion or merging tasks we use the RunConfig class to pass
        the desired function and job configuration.
        the apply() method is used to set resources like volumes, the with_secret() method adds secrets

        Most attributes are optional, if not specified a proper default value will be set

        examples::

            # config for local run emulation
            config = RunConfig(local=True)

            # config for using empty/default code
            config = RunConfig()

            # config for using .py/.ipynb file with image and extra package requirements
            config = RunConfig("mycode.py", image="mlrun/mlrun", requirements=["spacy"])

            # config for using function object
            function = mlrun.import_function("hub://some-function")
            config = RunConfig(function)

        :param function:    this can be function uri or function object or path to function code (.py/.ipynb)
                            or a :py:class:`~mlrun.runtimes.function_reference.FunctionReference`
                            the function define the code, dependencies, and resources
        :param local:       use True to simulate local job run or mock service
        :param image:       function container image
        :param kind:        function runtime kind (job, serving, spark, ..), required when function points to code
        :param handler:     the function handler to execute (for jobs or nuclio)
        :param parameters:  job parameters
        :param watch:       in batch jobs will wait for the job completion and print job logs to the console.
                            Default (None) is True.
        :param owner:       job owner
        :param credentials: job credentials
        :param code:        function source code (as string)
        :param requirements: python requirements file path or list of packages
        :param extra_spec:  additional dict with function spec fields/values to add to the function
        :param auth_info:   authentication info. *For internal use* when running on server
        """
        self._function = None
        self._modifiers = []
        self.secret_sources = []

        self.function = function
        self.local = local
        self.image = image
        self.kind = kind
        self.handler = handler
        self.parameters = parameters or {}
        self.watch = True if watch is None else watch
        self.owner = owner
        self.credentials = credentials
        self.code = code or ""
        self.requirements = requirements
        self.extra_spec = extra_spec
        self.auth_info = auth_info

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, function):
        if function and not (
            isinstance(function, (str, FunctionReference)) or hasattr(function, "apply")
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "function must be a uri (string) or mlrun function object/reference"
            )
        self._function = function

    def apply(self, modifier):
        """apply a modifier to add/set function resources like volumes

        example::

            run_config.apply(mlrun.platforms.auto_mount())
        """
        self._modifiers.append(modifier)
        return self

    def with_secret(self, kind, source):
        """register a secrets source (file, env or dict)

        read secrets from a source provider to be used in jobs, example::

            run_config.with_secrets('file', 'file.txt')
            run_config.with_secrets('inline', {'key': 'val'})
            run_config.with_secrets('env', 'ENV1,ENV2')
            run_config.with_secrets('vault', ['secret1', 'secret2'...])

        :param kind:   secret type (file, inline, env, vault)
        :param source: secret data or link (see example)

        :returns: This (self) object
        """

        self.secret_sources.append({"kind": kind, "source": source})
        return self

    def to_function(self, default_kind=None, default_image=None):
        """internal, generate function object"""
        if isinstance(self.function, FunctionReference):
            if self.code:
                self.function.code = self.code
            if self.requirements:
                self.function.requirements = self.requirements
            if self.extra_spec:
                self.function.spec = self.extra_spec
            function = self.function.to_function(
                default_kind, self.image or default_image
            )
        elif hasattr(self.function, "apply"):
            function = copy(self.function)
            if self.code:
                function.with_code(body=self.code)
            if self.requirements:
                self.function.with_requirements(self.requirements)
            if self.extra_spec:
                self.function = enrich_function_from_dict(
                    self.function, self.extra_spec
                )
            function.spec.image = function.spec.image or self.image or default_image
        else:
            function = FunctionReference(
                self.function,
                image=self.image,
                kind=self.kind,
                code=self.code,
                requirements=self.requirements,
                spec=self.extra_spec,
            ).to_function(default_kind, default_image)

        if not function.is_deployed():
            # todo: handle build for job functions
            logger.warn("cannot run function, it must be built/deployed first")

        for modifier in self._modifiers:
            function.apply(modifier)
        function.metadata.credentials = self.credentials
        return function

    def copy(self):
        return copy(self)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import typing
import warnings
from collections import namedtuple

import mlrun.common.secrets
import mlrun.errors
from mlrun.config import config
from mlrun.config import config as mlconf
from mlrun.errors import MLRunInvalidArgumentError
from mlrun.platforms.iguazio import v3io_to_vol
from mlrun.utils import logger

if typing.TYPE_CHECKING:
    from mlrun.runtimes import KubeResource


VolumeMount = namedtuple("Mount", ["path", "sub_path"])


def v3io_cred(
    api: str = "",
    user: str = "",
    access_key: str = "",
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to copy local v3io env vars to container

    Usage::

        train = train_op(...)
        train.apply(use_v3io_cred())
    """

    def _use_v3io_cred(runtime: "KubeResource"):
        web_api = api or os.environ.get("V3IO_API") or mlconf.v3io_api
        _user = user or os.environ.get("V3IO_USERNAME")
        _access_key = access_key or os.environ.get("V3IO_ACCESS_KEY")
        v3io_framesd = mlconf.v3io_framesd or os.environ.get("V3IO_FRAMESD")

        runtime.set_envs(
            {
                "V3IO_API": web_api,
                "V3IO_USERNAME": _user,
                "V3IO_ACCESS_KEY": _access_key,
                "V3IO_FRAMESD": v3io_framesd,
            },
        )

        return runtime

    return _use_v3io_cred


def mount_v3io(
    name: str = "v3io",
    remote: str = "",
    access_key: str = "",
    user: str = "",
    secret: typing.Optional[str] = None,
    volume_mounts: typing.Optional[list[VolumeMount]] = None,
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """Modifier function to apply to a Container Op to volume mount a v3io path

    :param name: the volume name
    :param remote: the v3io path to use for the volume (~/ prefix will be replaced with /users/<username>/)
    :param access_key: the access key used to auth against v3io (default: V3IO_ACCESS_KEY env var)
    :param user: the username used to auth against v3io (default: V3IO_USERNAME env var)
    :param secret: k8s secret name for the username and access key
    :param volume_mounts: list of VolumeMount; if empty, defaults to mounting /v3io and /User
    """
    volume_mounts, user = _enrich_and_validate_v3io_mounts(
        remote=remote,
        volume_mounts=volume_mounts,
        user=user,
    )

    def _attach_volume_mounts_and_creds(runtime: "KubeResource"):
        vol = v3io_to_vol(name, remote, access_key, user, secret=secret)
        runtime.spec.with_volumes(vol)

        for volume_mount in volume_mounts:
            runtime.spec.with_volume_mounts(
                {
                    "mountPath": volume_mount.path,
                    "name": name,
                    "subPath": volume_mount.sub_path,
                }
            )

        if not secret:
            runtime = v3io_cred(access_key=access_key, user=user)(runtime)
        return runtime

    return _attach_volume_mounts_and_creds


def mount_spark_conf() -> typing.Callable[["KubeResource"], "KubeResource"]:
    """Modifier function to mount Spark configuration."""

    def _mount_spark(runtime: "KubeResource"):
        runtime.spec.with_volume_mounts(
            {
                "mountPath": "/etc/config/spark",
                "name": "spark-master-config",
            }
        )
        return runtime

    return _mount_spark


def mount_v3iod(
    namespace: str, v3io_config_configmap: str
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """Modifier function to mount v3iod configuration."""

    def _mount_v3iod(runtime: "KubeResource"):
        def add_vol(name, mount_path, host_path):
            runtime.spec.with_volumes(
                {
                    "name": name,
                    "hostPath": {
                        "path": host_path,
                        "type": "",
                    },
                }
            )
            runtime.spec.with_volume_mounts(
                {
                    "mountPath": mount_path,
                    "name": name,
                }
            )

        add_vol(
            name="shm",
            mount_path="/dev/shm",
            host_path=f"/var/run/iguazio/dayman-shm/{namespace}",
        )
        add_vol(
            name="v3iod-comm",
            mount_path="/var/run/iguazio/dayman",
            host_path="/var/run/iguazio/dayman/" + namespace,
        )

        # Add daemon-health and v3io-config volumes
        runtime.spec.with_volumes(
            [
                {
                    "name": "daemon-health",
                    "emptyDir": {},
                },
                {
                    "name": "v3io-config",
                    "configMap": {
                        "name": v3io_config_configmap,
                        "defaultMode": 420,
                    },
                },
            ]
        )

        # Add volume mounts
        runtime.spec.with_volume_mounts(
            [
                {
                    "mountPath": "/var/run/iguazio/daemon_health",
                    "name": "daemon-health",
                },
                {
                    "mountPath": "/etc/config/v3io",
                    "name": "v3io-config",
                },
            ]
        )

        # Add environment variables
        runtime.set_envs(
            {
                "CURRENT_NODE_IP": {
                    "valueFrom": {
                        "fieldRef": {
                            "apiVersion": "v1",
                            "fieldPath": "status.hostIP",
                        }
                    },
                },
                "IGZ_DATA_CONFIG_FILE": "/igz/java/conf/v3io.conf",
            }
        )

        return runtime

    return _mount_v3iod


def mount_s3(
    secret_name: typing.Optional[str] = None,
    aws_access_key: str = "",
    aws_secret_key: str = "",
    endpoint_url: typing.Optional[str] = None,
    prefix: str = "",
    aws_region: typing.Optional[str] = None,
    non_anonymous: bool = False,
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """Modifier function to add s3 env vars or secrets to container

    :param secret_name: Kubernetes secret name for credentials
    :param aws_access_key: AWS_ACCESS_KEY_ID value (default: env variable)
    :param aws_secret_key: AWS_SECRET_ACCESS_KEY value (default: env variable)
    :param endpoint_url: s3 endpoint address (for non-AWS s3)
    :param prefix: prefix to add before the env var name (for multiple s3 data stores)
    :param aws_region: Amazon region
    :param non_anonymous: use non-anonymous connection even if no credentials are provided
            (for authenticating externally, such as through IAM instance-roles)

    """

    if secret_name and (aws_access_key or aws_secret_key):
        raise MLRunInvalidArgumentError(
            "Can use k8s_secret for credentials or specify them (aws_access_key, aws_secret_key) not both."
        )

    if not secret_name and (
        aws_access_key
        or os.environ.get(prefix + "AWS_ACCESS_KEY_ID")
        or aws_secret_key
        or os.environ.get(prefix + "AWS_SECRET_ACCESS_KEY")
    ):
        logger.warning(
            "It is recommended to use k8s secret (specify secret_name), "
            "specifying aws_access_key/aws_secret_key directly is unsafe."
        )

    def _use_s3_cred(runtime: "KubeResource"):
        _access_key = aws_access_key or os.environ.get(prefix + "AWS_ACCESS_KEY_ID")
        _secret_key = aws_secret_key or os.environ.get(prefix + "AWS_SECRET_ACCESS_KEY")

        # Check for endpoint URL with backward compatibility
        _endpoint_url = endpoint_url or os.environ.get(prefix + "AWS_ENDPOINT_URL_S3")
        if not _endpoint_url:
            # Check for deprecated environment variable
            _endpoint_url = os.environ.get(prefix + "S3_ENDPOINT_URL")
            if _endpoint_url:
                warnings.warn(
                    "S3_ENDPOINT_URL is deprecated in 1.10.0 and will be removed in 1.12.0, "
                    "use AWS_ENDPOINT_URL_S3 instead.",
                    # TODO: Remove this in 1.12.0
                    FutureWarning,
                )

        if _endpoint_url:
            runtime.set_env(prefix + "AWS_ENDPOINT_URL_S3", _endpoint_url)
        if aws_region:
            runtime.set_env(prefix + "AWS_REGION", aws_region)
        if non_anonymous:
            runtime.set_env(prefix + "S3_NON_ANONYMOUS", "true")

        if secret_name:
            runtime.set_envs(
                {
                    f"{prefix}AWS_ACCESS_KEY_ID": {
                        "valueFrom": {
                            "secretKeyRef": {
                                "name": secret_name,
                                "key": "AWS_ACCESS_KEY_ID",
                            }
                        }
                    },
                    f"{prefix}AWS_SECRET_ACCESS_KEY": {
                        "valueFrom": {
                            "secretKeyRef": {
                                "name": secret_name,
                                "key": "AWS_SECRET_ACCESS_KEY",
                            }
                        },
                    },
                }
            )
        else:
            runtime.set_envs(
                {
                    f"{prefix}AWS_ACCESS_KEY_ID": _access_key,
                    f"{prefix}AWS_SECRET_ACCESS_KEY": _secret_key,
                },
            )

        return runtime

    return _use_s3_cred


def mount_pvc(
    pvc_name: typing.Optional[str] = None,
    volume_name: str = "pipeline",
    volume_mount_path: str = "/mnt/pipeline",
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to mount a PVC volume in the container, simplifying volume and volume mount addition.

    Usage::

        train = train_op(...)
        train.apply(mount_pvc("claim-name", "pipeline", "/mnt/pipeline"))
    """
    if not pvc_name:
        # Try to get the PVC mount configuration from the environment variable
        if "MLRUN_PVC_MOUNT" in os.environ:
            mount = os.environ.get("MLRUN_PVC_MOUNT")
            items = mount.split(":")
            if len(items) != 2:
                raise MLRunInvalidArgumentError(
                    "MLRUN_PVC_MOUNT should include <pvc-name>:<mount-path>"
                )
            pvc_name = items[0]
            volume_mount_path = items[1]

    if not pvc_name:
        # The PVC name is still not set, raise an error
        raise MLRunInvalidArgumentError(
            "No PVC name: use the pvc_name parameter or configure the MLRUN_PVC_MOUNT environment variable"
        )

    def _mount_pvc(runtime: "KubeResource"):
        local_pvc = {"claimName": pvc_name}

        runtime.spec.with_volumes(
            [
                {
                    "name": volume_name,
                    "persistentVolumeClaim": local_pvc,
                }
            ]
        )
        runtime.spec.with_volume_mounts(
            {
                "mountPath": volume_mount_path,
                "name": volume_name,
            }
        )

        return runtime

    return _mount_pvc


def auto_mount(
    pvc_name: str = "",
    volume_mount_path: str = "",
    volume_name: typing.Optional[str] = None,
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """Choose the mount based on env variables and params

    Volume will be selected by the following order:

    - k8s PVC volume when both pvc_name and volume_mount_path are set
    - k8s PVC volume when env var is set: MLRUN_PVC_MOUNT=<pvc-name>:<mount-path>
    - k8s PVC volume if it's configured as the auto mount type
    - iguazio v3io volume when V3IO_ACCESS_KEY and V3IO_USERNAME env vars are set

    """
    if pvc_name and volume_mount_path:
        return mount_pvc(
            pvc_name=pvc_name,
            volume_mount_path=volume_mount_path,
            volume_name=volume_name or "shared-persistency",
        )
    if "MLRUN_PVC_MOUNT" in os.environ:
        return mount_pvc(
            volume_name=volume_name or "shared-persistency",
        )
    # In the case of CE when working remotely, no env variables will be defined but auto-mount
    # parameters may still be declared - use them in that case.
    if config.storage.auto_mount_type == "pvc":
        return mount_pvc(**config.get_storage_auto_mount_params())
    if "V3IO_ACCESS_KEY" in os.environ:
        return mount_v3io(name=volume_name or "v3io")

    raise ValueError("Failed to auto mount, need to set env vars")


def mount_secret(
    secret_name: str,
    mount_path: str,
    volume_name: str = "secret",
    items: typing.Optional[list[dict]] = None,
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to mount a Kubernetes secret as file(s).

    :param secret_name: Kubernetes secret name
    :param mount_path: Path inside the container to mount
    :param volume_name: Unique volume name
    :param items:        If unspecified, each key-value pair in the Data field
                         of the referenced Secret will be projected into the
                         volume as a file whose name is the key and content is
                         the value.
                         If specified, the listed keys will be projected into
                         the specified paths, and unlisted keys will not be
                         present."""

    if secret_name:
        mlrun.common.secrets.validate_not_forbidden_secret(secret_name.strip())

    def _mount_secret(runtime: "KubeResource"):
        # Define the secret volume source
        secret_volume_source = {
            "secretName": secret_name,
            "items": items,
        }

        # Add the secret volume
        runtime.spec.with_volumes(
            {
                "name": volume_name,
                "secret": secret_volume_source,
            }
        )

        # Add the volume mount
        runtime.spec.with_volume_mounts(
            {
                "mountPath": mount_path,
                "name": volume_name,
            }
        )

        return runtime

    return _mount_secret


def mount_configmap(
    configmap_name: str,
    mount_path: str,
    volume_name: str = "configmap",
    items: typing.Optional[list[dict]] = None,
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to mount a Kubernetes ConfigMap as file(s).

    :param configmap_name: Kubernetes ConfigMap name
    :param mount_path: Path inside the container to mount
    :param volume_name: Unique volume name
    :param items:           If unspecified, each key-value pair in the Data field
                            of the referenced Configmap will be projected into the
                            volume as a file whose name is the key and content is
                            the value.
                            If specified, the listed keys will be projected into
                            the specified paths, and unlisted keys will not be
                            present."""

    def _mount_configmap(runtime: "KubeResource"):
        # Construct the configMap dictionary
        config_map_dict = {
            "name": configmap_name,
        }
        if items is not None:
            config_map_dict["items"] = items

        vol = {
            "name": volume_name,
            "configMap": config_map_dict,
        }

        runtime.spec.with_volumes(vol)
        runtime.spec.with_volume_mounts(
            {
                "mountPath": mount_path,
                "name": volume_name,
            }
        )

        return runtime

    return _mount_configmap


def mount_hostpath(
    host_path: str,
    mount_path: str,
    volume_name: str = "hostpath",
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to mount a host path inside a Kubernetes container.

    :param host_path: Host path on the node to be mounted.
    :param mount_path: Path inside the container where the volume will be mounted.
    :param volume_name: Unique name for the volume.
    """

    def _mount_hostpath(runtime: "KubeResource") -> "KubeResource":
        runtime.spec.with_volumes(
            {
                "name": volume_name,
                "hostPath": {
                    "path": host_path,
                    "type": "",
                },
            }
        )
        runtime.spec.with_volume_mounts(
            {
                "mountPath": mount_path,
                "name": volume_name,
            }
        )

        return runtime

    return _mount_hostpath


def set_env_variables(
    env_vars_dict: typing.Optional[dict[str, str]] = None, **kwargs
) -> typing.Callable[["KubeResource"], "KubeResource"]:
    """
    Modifier function to apply a set of environment variables to a runtime. Variables may be passed
    as either a dictionary of name-value pairs, or as arguments to the function.
    See `KubeResource.apply` for more information on modifiers.

    Usage::

        function.apply(set_env_variables({"ENV1": "value1", "ENV2": "value2"}))
        or
        function.apply(set_env_variables(ENV1=value1, ENV2=value2))

    :param env_vars_dict: dictionary of environment variables
    :param kwargs: environment variables passed as arguments
    """

    env_data = env_vars_dict.copy() if env_vars_dict else {}
    for key, value in kwargs.items():
        env_data[key] = value

    def _set_env_variables(runtime: "KubeResource"):
        runtime.set_envs(env_data)

        return runtime

    return _set_env_variables


def _enrich_and_validate_v3io_mounts(
    remote: str = "",
    volume_mounts: typing.Optional[list[VolumeMount]] = None,
    user: str = "",
) -> tuple[list[VolumeMount], str]:
    if volume_mounts is None:
        volume_mounts = []
    if remote and not volume_mounts:
        raise MLRunInvalidArgumentError(
            "volume_mounts must be specified when remote is given"
        )

    # Empty remote & volume_mounts defaults are volume mounts of /v3io and /User
    if not remote and not volume_mounts:
        user = _resolve_mount_user(user)
        if not user:
            raise MLRunInvalidArgumentError(
                "user name/env must be specified when using empty remote and volume_mounts"
            )
        volume_mounts = [
            VolumeMount(path="/v3io", sub_path=""),
            VolumeMount(path="/User", sub_path="users/" + user),
        ]

    if not isinstance(volume_mounts, list) and any(
        [not isinstance(x, VolumeMount) for x in volume_mounts]
    ):
        raise TypeError("mounts should be a list of Mount")

    return volume_mounts, user


def _resolve_mount_user(user: typing.Optional[str] = None):
    return user or os.environ.get("V3IO_USERNAME")



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Final, Optional, Protocol, Union, cast

import numpy as np
from pandas import DataFrame, Series

import mlrun.artifacts
import mlrun.common.model_monitoring.helpers
import mlrun.model_monitoring.applications.context as mm_context
import mlrun.model_monitoring.applications.results as mm_results
import mlrun.model_monitoring.features_drift_table as mm_drift_table
from mlrun.common.schemas.model_monitoring.constants import (
    HistogramDataDriftApplicationConstants,
    ResultKindApp,
    ResultStatusApp,
    StatsKind,
)
from mlrun.model_monitoring.applications import (
    ModelMonitoringApplicationBase,
)
from mlrun.model_monitoring.metrics.histogram_distance import (
    HellingerDistance,
    HistogramDistanceMetric,
    KullbackLeiblerDivergence,
    TotalVarianceDistance,
)


class InvalidMetricValueError(ValueError):
    pass


class InvalidThresholdValueError(ValueError):
    pass


class ValueClassifier(Protocol):
    def value_to_status(self, value: float) -> ResultStatusApp: ...


@dataclass
class DataDriftClassifier:
    """
    Classify data drift numeric values into categorical status.
    """

    potential: float = 0.5
    detected: float = 0.7

    def __post_init__(self) -> None:
        """Catch erroneous threshold values"""
        if not 0 < self.potential < self.detected < 1:
            raise InvalidThresholdValueError(
                "The provided thresholds do not comply with the rules"
            )

    def value_to_status(self, value: float) -> ResultStatusApp:
        """
        Translate the numeric value into status category.

        :param value: The numeric value of the data drift metric, between 0 and 1.
        :returns:     `ResultStatusApp` according to the classification.
        """
        if value > 1 or value < 0:
            raise InvalidMetricValueError(
                f"{value = } is invalid, must be in the range [0, 1]."
            )
        if value >= self.detected:
            return ResultStatusApp.detected
        if value >= self.potential:
            return ResultStatusApp.potential_detection
        return ResultStatusApp.no_detection


class HistogramDataDriftApplication(ModelMonitoringApplicationBase):
    """
    MLRun's default data drift application for model monitoring.

    The application expects tabular numerical data, and calculates three metrics over the shared features' histograms.
    The metrics are calculated on features that have reference data from the training dataset. When there is no
    reference data (`feature_stats`), this application send a warning log and does nothing.
    The three metrics are:

    * Hellinger distance.
    * Total variance distance.
    * Kullback-Leibler divergence.

    Each metric is calculated over all the features individually and the mean is taken as the metric value.
    The average of Hellinger and total variance distance is taken as the result.

    The application can log two artifacts (disabled by default due to performance issues):

    * JSON with the general drift value per feature.
    * Plotly table with the various metrics and histograms per feature.

    This application is deployed by default when calling
    :py:func:`~mlrun.projects.MlrunProject.enable_model_monitoring`.
    To avoid it, pass :code:`deploy_histogram_data_drift_app=False`.

    If you want to change the application defaults, such as the classifier or which artifacts to produce, you
    need to inherit from this class and deploy it as any other model monitoring application.
    Please make sure to keep the default application name. This ensures that the full functionality of the application,
    including the statistics view in the UI, is available.
    """

    NAME: Final[str] = HistogramDataDriftApplicationConstants.NAME

    _REQUIRED_METRICS = {HellingerDistance, TotalVarianceDistance}
    _STATS_TYPES: tuple[StatsKind, StatsKind] = (
        StatsKind.CURRENT_STATS,
        StatsKind.DRIFT_MEASURES,
    )

    metrics: list[type[HistogramDistanceMetric]] = [
        HellingerDistance,
        KullbackLeiblerDivergence,
        TotalVarianceDistance,
    ]

    def __init__(
        self,
        value_classifier: Optional[ValueClassifier] = None,
        produce_json_artifact: bool = False,
        produce_plotly_artifact: bool = False,
    ) -> None:
        """
        :param value_classifier:        Classifier object that adheres to the :py:class:`~ValueClassifier` protocol.
                                        If not provided, the default :py:class:`~DataDriftClassifier` is used.
        :param produce_json_artifact:   Whether to produce the JSON artifact or not, ``False`` by default.
        :param produce_plotly_artifact: Whether to produce the Plotly artifact or not, ``False`` by default.
        """
        self._value_classifier = value_classifier or DataDriftClassifier()
        assert self._REQUIRED_METRICS <= set(
            self.metrics
        ), "TVD and Hellinger distance are required for the general data drift result"

        self._produce_json_artifact = produce_json_artifact
        self._produce_plotly_artifact = produce_plotly_artifact

    def _compute_metrics_per_feature(
        self, monitoring_context: mm_context.MonitoringApplicationContext
    ) -> DataFrame:
        """Compute the metrics for the different features and labels"""
        metrics_per_feature = DataFrame(
            columns=[metric_class.NAME for metric_class in self.metrics]
        )
        feature_stats = monitoring_context.dict_to_histogram(
            monitoring_context.feature_stats
        )
        sample_df_stats = monitoring_context.dict_to_histogram(
            monitoring_context.sample_df_stats
        )
        for feature_name in feature_stats:
            sample_hist = np.asarray(sample_df_stats[feature_name])
            reference_hist = np.asarray(feature_stats[feature_name])
            monitoring_context.logger.info(
                "Computing metrics for feature", feature_name=feature_name
            )
            metrics_per_feature.loc[feature_name] = {  # pyright: ignore[reportCallIssue,reportArgumentType]
                metric.NAME: metric(
                    distrib_t=sample_hist, distrib_u=reference_hist
                ).compute()
                for metric in self.metrics
            }
        monitoring_context.logger.info("Finished computing the metrics")

        return metrics_per_feature

    def _get_general_drift_result(
        self, metrics: list[mm_results.ModelMonitoringApplicationMetric]
    ) -> mm_results.ModelMonitoringApplicationResult:
        """Get the general drift result from the metrics list"""
        value = cast(
            float,
            np.mean(
                [
                    metric.value
                    for metric in metrics
                    if metric.name
                    in [
                        f"{HellingerDistance.NAME}_mean",
                        f"{TotalVarianceDistance.NAME}_mean",
                    ]
                ]
            ),
        )

        status = self._value_classifier.value_to_status(value)

        return mm_results.ModelMonitoringApplicationResult(
            name=HistogramDataDriftApplicationConstants.GENERAL_RESULT_NAME,
            value=value,
            kind=ResultKindApp.data_drift,
            status=status,
        )

    @staticmethod
    def _get_metrics(
        metrics_per_feature: DataFrame,
    ) -> list[mm_results.ModelMonitoringApplicationMetric]:
        """Average the metrics over the features and add the status"""
        metrics: list[mm_results.ModelMonitoringApplicationMetric] = []

        metrics_mean = metrics_per_feature.mean().to_dict()

        for name, value in metrics_mean.items():
            metrics.append(
                mm_results.ModelMonitoringApplicationMetric(
                    name=f"{name}_mean",
                    value=value,
                )
            )

        return metrics

    @staticmethod
    def _get_stats(
        metrics: list[mm_results.ModelMonitoringApplicationMetric],
        metrics_per_feature: DataFrame,
        monitoring_context: mm_context.MonitoringApplicationContext,
    ) -> list[mm_results._ModelMonitoringApplicationStats]:
        """
        Return a list of the statistics.

        :param metrics: the calculated metrics
        :param metrics_per_feature: metric calculated per feature
        :param monitoring_context:  context object for current monitoring application
        :returns: list of mm_results._ModelMonitoringApplicationStats for histogram data drift application
        """
        stats = []
        for stats_type in HistogramDataDriftApplication._STATS_TYPES:
            stats.append(
                mm_results._ModelMonitoringApplicationStats(
                    name=stats_type,
                    stats=metrics_per_feature.T.to_dict()
                    | {metric.name: metric.value for metric in metrics}
                    if stats_type == StatsKind.DRIFT_MEASURES
                    else monitoring_context.sample_df_stats,
                    timestamp=monitoring_context.end_infer_time.isoformat(
                        sep=" ", timespec="microseconds"
                    ),
                )
            )
        return stats

    @staticmethod
    def _get_shared_features_sample_stats(
        monitoring_context: mm_context.MonitoringApplicationContext,
    ) -> mlrun.common.model_monitoring.helpers.FeatureStats:
        """
        Filter out features without reference data in `feature_stats`, e.g. `timestamp`.
        """
        return mlrun.common.model_monitoring.helpers.FeatureStats(
            {
                key: monitoring_context.sample_df_stats[key]
                for key in monitoring_context.feature_stats
            }
        )

    @staticmethod
    def _log_json_artifact(
        drift_per_feature_values: Series,
        monitoring_context: mm_context.MonitoringApplicationContext,
    ) -> None:
        """Log the drift values as a JSON artifact"""
        monitoring_context.logger.debug("Logging drift value per feature JSON artifact")
        monitoring_context.log_artifact(
            mlrun.artifacts.Artifact(
                body=drift_per_feature_values.to_json(),
                format="json",
                key="features_drift_results",
            )
        )
        monitoring_context.logger.debug("Logged JSON artifact successfully")

    def _log_plotly_table_artifact(
        self,
        sample_set_statistics: mlrun.common.model_monitoring.helpers.FeatureStats,
        inputs_statistics: mlrun.common.model_monitoring.helpers.FeatureStats,
        metrics_per_feature: DataFrame,
        drift_per_feature_values: Series,
        monitoring_context: mm_context.MonitoringApplicationContext,
    ) -> None:
        """Log the Plotly drift table artifact"""
        monitoring_context.logger.debug(
            "Feature stats",
            sample_set_statistics=sample_set_statistics,
            inputs_statistics=inputs_statistics,
        )

        monitoring_context.logger.debug("Computing drift results per feature")
        drift_results = {
            cast(str, key): (self._value_classifier.value_to_status(value), value)
            for key, value in drift_per_feature_values.items()
        }
        monitoring_context.logger.debug("Producing plotly artifact")
        artifact = mm_drift_table.FeaturesDriftTablePlot().produce(
            sample_set_statistics=sample_set_statistics,
            inputs_statistics=inputs_statistics,
            metrics=metrics_per_feature.T.to_dict(),  # pyright: ignore[reportArgumentType]
            drift_results=drift_results,
        )
        monitoring_context.logger.debug("Logging plotly artifact")
        monitoring_context.log_artifact(artifact)
        monitoring_context.logger.debug("Logged plotly artifact successfully")

    def _log_drift_artifacts(
        self,
        monitoring_context: mm_context.MonitoringApplicationContext,
        metrics_per_feature: DataFrame,
    ) -> None:
        """Log JSON and Plotly drift data per feature artifacts"""
        if not self._produce_json_artifact and not self._produce_plotly_artifact:
            return

        drift_per_feature_values = metrics_per_feature[
            [HellingerDistance.NAME, TotalVarianceDistance.NAME]
        ].mean(axis=1)

        if self._produce_json_artifact:
            self._log_json_artifact(drift_per_feature_values, monitoring_context)

        if self._produce_plotly_artifact:
            self._log_plotly_table_artifact(
                sample_set_statistics=self._get_shared_features_sample_stats(
                    monitoring_context
                ),
                inputs_statistics=monitoring_context.feature_stats,
                metrics_per_feature=metrics_per_feature,
                drift_per_feature_values=drift_per_feature_values,
                monitoring_context=monitoring_context,
            )

    def do_tracking(
        self, monitoring_context: mm_context.MonitoringApplicationContext
    ) -> list[
        Union[
            mm_results.ModelMonitoringApplicationResult,
            mm_results.ModelMonitoringApplicationMetric,
            mm_results._ModelMonitoringApplicationStats,
        ]
    ]:
        """
        Calculate and return the data drift metrics, averaged over the features.
        """
        monitoring_context.logger.debug("Starting to run the application")
        if not monitoring_context.feature_stats:
            monitoring_context.logger.warning(
                "No feature statistics found, skipping the application. \n"
                "In order to run the application, training set must be provided when logging the model."
            )
            return []
        metrics_per_feature = self._compute_metrics_per_feature(
            monitoring_context=monitoring_context
        )
        monitoring_context.logger.debug("Saving artifacts")
        self._log_drift_artifacts(
            monitoring_context=monitoring_context,
            metrics_per_feature=metrics_per_feature,
        )
        monitoring_context.logger.debug("Computing average per metric")
        metrics = self._get_metrics(metrics_per_feature)
        result = self._get_general_drift_result(metrics=metrics)
        stats = self._get_stats(
            metrics=metrics,
            monitoring_context=monitoring_context,
            metrics_per_feature=metrics_per_feature,
        )
        metrics_result_and_stats = metrics + [result] + stats
        monitoring_context.logger.debug(
            "Finished running the application", results=metrics_result_and_stats
        )
        return metrics_result_and_stats



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._archiver import ArchiveSupportedFormat
from ._formatter import StructFileSupportedFormat
from ._pickler import Pickler
from ._supported_format import SupportedFormat
from .log_hint_utils import LogHintKey, LogHintUtils
from .type_hint_utils import TypeHintUtils

# The default pickle module to use for pickling objects:
DEFAULT_PICKLE_MODULE = "cloudpickle"
# The default archive format to use for archiving directories:
DEFAULT_ARCHIVE_FORMAT = ArchiveSupportedFormat.ZIP
# The default struct file format to use for savings python struct objects (dictionaries and lists):
DEFAULT_STRUCT_FILE_FORMAT = StructFileSupportedFormat.JSON


class ArtifactType:
    """
    Possible artifact types to pack objects as and log using a `mlrun.Packager`.
    """

    OBJECT = "object"
    PATH = "path"
    FILE = "file"
    DATASET = "dataset"
    MODEL = "model"
    PLOT = "plot"
    RESULT = "result"


class DatasetFileFormat:
    """
    All file format for logging objects as `DatasetArtifact`.
    """

    CSV = "csv"
    PARQUET = "parquet"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing

import pydantic.v1

import mlrun.common.types

from .common import ImageBuilder
from .object import ObjectKind, ObjectStatus


class ProjectMetadata(pydantic.v1.BaseModel):
    name: str
    created: typing.Optional[datetime.datetime] = None
    labels: typing.Optional[dict] = {}
    annotations: typing.Optional[dict] = {}

    class Config:
        extra = pydantic.v1.Extra.allow


class ProjectDesiredState(mlrun.common.types.StrEnum):
    online = "online"
    offline = "offline"
    archived = "archived"


class ProjectState(mlrun.common.types.StrEnum):
    unknown = "unknown"
    creating = "creating"
    deleting = "deleting"
    online = "online"
    offline = "offline"
    archived = "archived"

    @staticmethod
    def terminal_states():
        return [
            ProjectState.online,
            ProjectState.offline,
            ProjectState.archived,
        ]


class ProjectStatus(ObjectStatus):
    state: typing.Optional[ProjectState]


class ProjectSpec(pydantic.v1.BaseModel):
    description: typing.Optional[str] = None
    owner: typing.Optional[str] = None
    goals: typing.Optional[str] = None
    params: typing.Optional[dict] = {}
    functions: typing.Optional[list] = []
    workflows: typing.Optional[list] = []
    artifacts: typing.Optional[list] = []
    artifact_path: typing.Optional[str] = None
    conda: typing.Optional[str] = None
    source: typing.Optional[str] = None
    subpath: typing.Optional[str] = None
    origin_url: typing.Optional[str] = None
    desired_state: typing.Optional[ProjectDesiredState] = ProjectDesiredState.online
    custom_packagers: typing.Optional[list[tuple[str, bool]]] = None
    default_image: typing.Optional[str] = None
    build: typing.Optional[ImageBuilder] = None
    default_function_node_selector: typing.Optional[dict] = {}

    class Config:
        extra = pydantic.v1.Extra.allow


class ProjectSpecOut(pydantic.v1.BaseModel):
    description: typing.Optional[str] = None
    owner: typing.Optional[str] = None
    goals: typing.Optional[str] = None
    params: typing.Optional[dict] = {}
    functions: typing.Optional[list] = []
    workflows: typing.Optional[list] = []
    artifacts: typing.Optional[list] = []
    artifact_path: typing.Optional[str] = None
    conda: typing.Optional[str] = None
    source: typing.Optional[str] = None
    subpath: typing.Optional[str] = None
    origin_url: typing.Optional[str] = None
    desired_state: typing.Optional[ProjectDesiredState] = ProjectDesiredState.online
    custom_packagers: typing.Optional[list[tuple[str, bool]]] = None
    default_image: typing.Optional[str] = None
    build: typing.Any = None
    default_function_node_selector: typing.Optional[dict] = {}

    class Config:
        extra = pydantic.v1.Extra.allow


class Project(pydantic.v1.BaseModel):
    kind: ObjectKind = pydantic.v1.Field(ObjectKind.project, const=True)
    metadata: ProjectMetadata
    spec: ProjectSpec = ProjectSpec()
    status: ObjectStatus = ObjectStatus()


# The reason we have a different schema for the response model is that we don't want to validate project.spec.build in
# the response as the validation was added late and there may be corrupted values in the DB.
class ProjectOut(pydantic.v1.BaseModel):
    kind: ObjectKind = pydantic.v1.Field(ObjectKind.project, const=True)
    metadata: ProjectMetadata
    spec: ProjectSpecOut = ProjectSpecOut()
    status: ObjectStatus = ObjectStatus()


class ProjectOwner(pydantic.v1.BaseModel):
    username: str
    access_key: str


class ProjectSummary(pydantic.v1.BaseModel):
    name: str
    files_count: int = 0
    feature_sets_count: int = 0
    models_count: int = 0
    runs_completed_recent_count: int = 0
    runs_failed_recent_count: int = 0
    runs_running_count: int = 0
    distinct_schedules_count: int = 0
    distinct_scheduled_jobs_pending_count: int = 0
    distinct_scheduled_pipelines_pending_count: int = 0
    pipelines_completed_recent_count: typing.Optional[int] = None
    pipelines_failed_recent_count: typing.Optional[int] = None
    pipelines_running_count: typing.Optional[int] = None
    updated: typing.Optional[datetime.datetime] = None
    endpoint_alerts_count: int = 0
    job_alerts_count: int = 0
    other_alerts_count: int = 0
    datasets_count: int = 0
    documents_count: int = 0
    llm_prompts_count: int = 0
    running_model_monitoring_functions: int = 0
    failed_model_monitoring_functions: int = 0
    real_time_model_endpoint_count: int = 0
    batch_model_endpoint_count: int = 0


class IguazioProject(pydantic.v1.BaseModel):
    data: dict


# The format query param controls the project type used:
# full - ProjectOut
# name_only - str
# summary - ProjectSummary
# leader - currently only IguazioProject supported
# The way pydantic handles typing.Union is that it takes the object and tries to coerce it to be the types of the
# union by the definition order. Therefore, we can't currently add generic dict for all leader formats, but we need
# to add a specific classes for them. it's frustrating but couldn't find other workaround, see:
# https://github.com/samuelcolvin/pydantic/issues/1423, https://github.com/samuelcolvin/pydantic/issues/619
ProjectOutput = typing.TypeVar(
    "ProjectOutput",
    ProjectOut,
    str,
    ProjectSummary,
    IguazioProject,
    tuple[str, datetime.datetime],
)


class ProjectsOutput(pydantic.v1.BaseModel):
    projects: list[ProjectOutput]


class ProjectSummariesOutput(pydantic.v1.BaseModel):
    project_summaries: list[ProjectSummary]
